try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.utils_base import *
from omfit_classes.sortedDict import *

import numpy as np
import warnings
import uncertainties

comment_ptrn = re.compile(r'^__comment.*__$')

def _debug():
    try:
        return int(os.environ.get('OMFIT_DEBUG','0'))
    except ValueError:
        return 0

def _split_arrays_function(name, value, offset=1, sparse_indexing=False):
    names=[]
    values=[]

    if isinstance(value,sparray):
        tmp=value.fortran_repr()
        for off,sub_value in zip(list(tmp.keys()),list(tmp.values())):
            if len(off):
                N,V=_split_arrays_function(name,sub_value,eval(off),sparse_indexing=True)
                names.extend(N)
                values.extend(V)
            else:
                names.append(name+'()')
                values.append(sub_value)
        return names,values

    for k,v in np.ndenumerate(value):
        k=np.atleast_1d(offset)*0+np.atleast_1d(k)
        if sparse_indexing and len(k)>1:
            k[1:]*=0
        k=tuple(k+offset)
        names.append('%s%s' % (name, re.sub(' ', '', re.sub(r',\)', ')', str(k)))))
        values.append([v])

    return names,values

def interpreter(orig, escaped_strings=True):
    """
    Parse string value in a fortran namelist format

    NOTE: for strings that are arrays of elements one may use the following notation:
    >> lines = '1 1.2 2.3 4.5 8*5.6'
    >> values = []
    >> for item in re.split('[ |\t]+', line.strip()):
    >>     values.extend(tolist(namelist.interpreter(item)))

    :param orig: string value element in a fortran namelist format

    :param escaped_strings: do strings follow proper escaping

    :return: parsed namelist element
    """

    # originally everything is in form of string but already split
    def eval_(a):
        try:
            return int(a)
        except Exception:
            try:
                return float(a)
            except Exception:
                return ast.literal_eval(a)

    out = orig
    if not isinstance(out, str):
        out = repr(orig)

    try:
        # fist try a number or a string that has been correctly escaped
        # remove fortran namelist escaping
        if escaped_strings:
            orig_ = re.sub(r'\\([!&$])', r'\1', orig)
            # fortran does not escape newlines
            if escaped_strings == 'fortran':
                orig_ = re.sub(r'\\', r'\\\\', orig_)
        else:
            orig_ = orig
        out = eval_(orig_)
        if isinstance(out, tuple):
            return complex(out[0], out[1])
    except Exception:
        try:
            if '+/-' in orig:
                out = uncertainties.ufloat_fromstr(orig)
            elif orig == 'nan':
                out = np.nan
            elif orig == 'inf':
                out = np.inf
            elif re.match(r'\- ?inf', orig):
                out = -np.inf
            else:
                raise
        except Exception:
            try:
                # then try a number saved with float format 0.d0
                tmp = re.sub(r'([0-9]+|[0-9]*(?:[0-9]\.|\.[0-9])[0-9]*)[dD]([\-\+]*[0-9]+)', r'\1e\2', orig)
                # also remove +'s
                tmp = re.sub(r'\+', '', tmp)
                out = eval_(tmp)
            except (SyntaxError, ValueError):
                try:
                    # then check if it's one of those fortran repetition array
                    if re.match(r'([0-9]+)\*(.+)', tmp):
                        nrep = int(re.sub(r'([0-9]+)\*(.+)', r'\1', tmp))
                        val = re.sub(r'([0-9]+)\*(.+)', r'\2', tmp)
                        out = [eval_(val)] * nrep
                    else:
                        raise
                except Exception:
                    try:
                        # then try a boolean
                        tmp = re.sub(r'(?i)^\.true\.', '1', orig)
                        tmp = re.sub(r'(?i)^\.false\.', '0', tmp)
                        tmp = re.sub(r'(?i)^true', '1', tmp)
                        tmp = re.sub(r'(?i)^false', '0', tmp)
                        tmp = re.sub(r'(?i)^\.t', '1', tmp)
                        tmp = re.sub(r'(?i)^\.f', '0', tmp)
                        tmp = re.sub(r'(?i)^t', '1', tmp)
                        tmp = re.sub(r'(?i)^f', '0', tmp)
                        tmp = bool(eval_(tmp))
                        out = tmp
                    except (SyntaxError, ValueError):
                        pass
    return out

def array_encoder(orig, line='', level=0, separator_arrays=' ', compress_arrays=True, max_array_chars=79, escaped_strings=True):
    all=line
    tmps=list([encoder(orig,escaped_strings) for orig in np.atleast_1d(orig)])

    #compress repeted entries
    if compress_arrays and len(tmps)>0:
        tmpsc=[]
        c=-1
        kold=tmps[0]
        for k in tmps+[tmps[-1]+' ']:
            if kold==k:
                c+=1
            else:
                if c==0:
                    tmpsc.append(kold)
                else:
                    tmpsc.append(str(c+1)+'*'+kold)
                kold=k
                c=0
        tmps=tmpsc

    #wrap entries so that they do not take more than max_array_chars per lines
    for k,tmp in enumerate(tmps):
        if k!=(len(tmps)-1):
            tmp=tmp+separator_arrays
        if not max_array_chars or len(line)+len(tmp)<max_array_chars:
            all+=tmp
            line+=tmp
        else:
            line=' '*level+tmp
            all+='\n'+line
    return all

def encoder(orig, escaped_strings=True, dotBoolean=True, compress_arrays=True, max_array_chars=79):
    if isinstance(orig,(bool,np.bool_)):
        if dotBoolean:
            tmp='.'+str(orig)+'.'
        else:
            tmp=str(orig)
    elif escaped_strings and isinstance(orig,str):
        tmp=repr(orig)
        tmp=re.sub('([!&$])',r'\\\1',tmp)
        if escaped_strings=='fortran':
            tmp=re.sub(r'\\\\',r'\\',tmp)
    elif isinstance(orig,uncertainties.core.AffineScalarFunc):
        tmp=repr(orig)
    elif isinstance(orig,complex):
        tmp=repr((float(np.real(orig)),float(np.imag(orig))))
    elif isinstance(orig,tuple):
        tmp=array_encoder(np.array(list(orig)).flatten(),compress_arrays=compress_arrays,max_array_chars=max_array_chars,escaped_strings=escaped_strings)
    elif isinstance(orig,list):
        tmp=array_encoder(np.array(orig).flatten(),compress_arrays=compress_arrays,max_array_chars=max_array_chars,escaped_strings=escaped_strings)
    elif isinstance(orig,np.ndarray):
        tmp=array_encoder(orig.flatten(),compress_arrays=compress_arrays,max_array_chars=max_array_chars,escaped_strings=escaped_strings)
    elif orig is None:
        tmp=''
    elif orig == None:
        tmp=''
    else:
        tmp=str(orig)
    return tmp

class NamelistName(SortedDict):
    """Defines variables defined within a &XXX and / delimiter in a FOTRAN namelist"""
    def __init__(self, *args, **kw):
        SortedDict.__init__(self, *args, **kw)
        self.caseInsensitive = True
        self._collect = False
        self.collect_arrays = False
        self.index_offset = kw.get('index_offset', False)

    def _checkSetitem(self, key, value):
        '''
        Namelists can only handle strings as their dictionary keys
        This function will raise an exception during __setitem__ if that's not the case.
        '''
        if not isinstance(key,str):
            raise Exception('Namelist only allows strings as dictionary keys')
        else:
            return key, value

    def collectArrays(self, _dimensions=None, **input_dimensions):
        '''
        This function collects the multiple namelist arrays into a single one::

             collectArrays(**{'__default__':0,        # default value for whole namelist (use when no default is found)
                              'BCMOM':{               # options for specific entry in the namelist
                                  'default':3,        # a default value must be defined to perform math ops (automatically set by a=... )
                                  'shape':(30,30),    # this overrides automatic shape detection (automatically set by a(30,30)=...)
                                  'offset':(-10,-10), # this overrides automatic offset detection (automatically set to be the minimum of the offsets of the entries in all dimensions a(-10,-10)=...)
                                  'dtype':0}          # this overrides automatic type detection (automatically set to float if at least one float is found)
                              })
        '''
        collect = SortedDict(caseInsensitive=True)
        defaults = SortedDict(caseInsensitive=True)
        offsets = SortedDict(caseInsensitive=True)
        dtypes = SortedDict(caseInsensitive=True)
        outputs = SortedDict(caseInsensitive=True)
        shapes = SortedDict(caseInsensitive=True)
        dimensions = SortedDict(caseInsensitive=True)

        dimensions.update(input_dimensions)
        if isinstance(_dimensions, dict):
            dimensions.update(_dimensions)
        if not isinstance(self.collect_arrays, dict):
            self.collect_arrays = SortedDict(caseInsensitive=True)
        self.collect_arrays.update(dimensions)

        dimensions.update(self.collect_arrays)

        # find the arrays to collect
        for kid in list(self.keys()):
            if isinstance(self[kid], NamelistName):
                self[kid].collect_arrays = self.collect_arrays
                self[kid].collectArrays(dimensions, **input_dimensions)
                continue
            if re.match(r'.*\(.*\)', kid)  or kid in self.collect_arrays:
                if re.match(r'.*\(.*\)', kid):
                    var = re.sub(r'(.*)\((.*)\)', r'\1', kid)
                else:
                    var = kid
                # add to list of items to append
                collect.setdefault(var, [])

        # expand ranges
        for kid in list(self.keys()):
            mat = re.match(r'(.*)\((.*)\)', kid)
            if not mat:
                continue
            if self[kid] is None:
                printw('Assigning 0 length array to finite range: %s' % kid)
                continue
            sk = mat.groups()
            if len(sk) < 2:
                continue
            var, ind = sk
            if ':' not in ind:
                continue
            rnk_ranges = []
            ranks = ind.split(',')
            for ri, rank in enumerate(ranks):
                if ':' not in rank:
                    continue
                rnk_ranges.append(ri)
            if len(rnk_ranges) > 1:
                raise NotImplementedError(f'{kid}: must only give one dimension with a range')
            rng_ri = rnk_ranges[0]
            new_key = '{var}({prefix}{rg_ind}{postfix})'
            prefix = ','.join(ranks[:rng_ri])
            if rng_ri > 0:
                prefix = prefix + ','
            postfix = ''
            if rng_ri != len(ranks) - 1:
                postfix = postfix + ','
            postfix = postfix + ','.join(ranks[rng_ri + 1:])

            rngs = list(map(int, ranks[rng_ri].split(':')))
            rngs[1] = rngs[1] + 1
            if len(list(range(*rngs))) != len(self[kid]):
                raise ValueError('Length of range and values disagree for %s: %s; %s' % (kid, ranks[rng_ri], self[kid]))
            for ri, rang in enumerate(range(*rngs)):
                self[new_key.format(prefix=prefix, var=var, rg_ind=rang, postfix=postfix)] = self[kid][ri]
            del self[kid]

        # find the arrays to collect
        warn_msg = []
        for kid in list(self.keys()):
            if re.sub(r'(.*)\((.*)\)', r'\1', kid) in collect or kid in self.collect_arrays and self.collect_arrays[kid].get('sparray', True):
                if re.match(r'.*\(.*\)', kid):
                    try:
                        var = re.sub(r'(.*)\((.*)\)', r'\1', kid)
                        indexStr = re.sub(r'(.*)\((.*)\)', r'\2', kid)
                        index = np.array(list(map(int, indexStr.split(','))))
                    except ValueError:
                        continue
                else:
                    var = kid
                    indexStr = None
                    index = (1,)

                # types
                if var in dimensions and 'dtype' in dimensions[var]:
                    tmp = dimensions[var]['dtype']
                    if callable(tmp):
                        tmp = tmp(0)
                    dtypes[var] = tmp
                elif np.atleast_1d(self[kid]).dtype.kind.lower() in ['s', 'u']:
                    dtypes[var] = ''  # string
                elif np.atleast_1d(self[kid]).dtype.kind.lower() in ['f']:
                    dtypes[var] = 0.0  # float
                elif np.atleast_1d(self[kid]).dtype.kind.lower() in['b']:
                    dtypes[var] = False  # boolean
                elif var not in dtypes:
                    dtypes[var] = 0  # int

                # offsets
                if var not in offsets:
                    offsets[var] = index
                off = np.ones(np.shape(index))
                off[:len(offsets[var])] = offsets[var]
                offsets[var] = np.min([off, index], 0)
                offsets[var] = tuple(map(int, offsets[var].tolist()))

                # add to list of items to append
                collect[var].append(indexStr)

        # user defined shapes
        for var in list(collect.keys()):
            if var in dimensions and 'shape' in dimensions[var]:
                shapes[var] = tuple(dimensions[var]['shape'])
            else:
                shapes[var] = None

        # if there is something to collect
        if len(collect):
            # if there was a default as array update the offsets to be at maximum 1
            for var in list(collect.keys()):
                for k, indexStr in enumerate(collect[var]):
                    if indexStr is None:
                        if len(collect[var]) > 1 or offsets[var] != 0:
                            offsets[var] = np.min([offsets[var], np.ones(np.shape(offsets[var]))], 0)
                            offsets[var] = tuple(map(int, offsets[var].tolist()))
                        else:
                            del collect[var]
                            del offsets[var]
                            continue

            # user defined offsets
            for var in list(offsets.keys()):
                if var in dimensions and 'offset' in dimensions[var]:
                    offsets[var] = tuple(dimensions[var]['offset'])
                elif '__offset__' in dimensions:
                    offsets[var] = tuple([dimensions['__offset__']] * len(offsets[var]))
                else:
                    # equal offsets for all dimensions (reasonable?)
                    offsets[var] = tuple([min(offsets[var])] * len(offsets[var]))

            # defaults
            for var in list(offsets.keys()):
                if isinstance(dtypes[var], str):
                    defaults[var] = dimensions.get('__default__', '')
                else:
                    defaults[var] = dimensions.get('__default__', np.nan)
            for var in list(offsets.keys()):
                if var in dimensions and 'default' in dimensions[var]:
                    defaults[var] = dimensions[var]['default']

            # define sparse arrays
            for var in list(offsets.keys()):

                # handle arrays of strings
                if isinstance(dtypes[var], str):
                    # for the time being only handle var(1)=['a','b','c'] type string arrays
                    if len(offsets[var]) > 1:
                        continue

                    # figure out the shape
                    if shapes[var] is None:
                        shapes[var] = 0
                        for indexStr in collect[var]:
                            if indexStr is None:
                                defaults[var] = self[var]
                                continue
                            kid = '%s(%s)' % (var, indexStr)
                            shapes[var] = max([shapes[var], int(indexStr) + len(tolist(self[kid])) - 1])
                        shapes[var] = (shapes[var],)

                    # generate array
                    outputs[var] = [defaults[var]] * shapes[var][0]
                    for indexStr in collect[var]:
                        if indexStr is None:
                            continue
                        kid = '%s(%s)' % (var, indexStr)
                        for k, value in enumerate(tolist(self[kid])):
                            outputs[var][int(indexStr) - offsets[var][0] + k] = value

                # handle numerical arrays
                else:
                    outputs[var] = sparray(shapes[var], offset=offsets[var], default=defaults[var], dtype=dtypes[var], index_offset=True)
                    for indexStr in collect[var]:
                        if indexStr is None:
                            kid = var
                            indexStr = ','.join(list(map(str, offsets[var])))
                        else:
                            kid = '%s(%s)' % (var, indexStr)
                        outputs[var].fortran(indexStr, self[kid])
                    outputs[var].index_offset = self.index_offset
                    # keep it sparse if requested by user
                    if var in self.collect_arrays and self.collect_arrays[var].get('sparray', True):
                        pass

                    # dense if 1D or 2D and offsets start at 1
                    elif ((outputs[var].ndim == 1 and np.atleast_1d(outputs[var].offset)[0] == 1) or
                            (outputs[var].ndim == 2 and np.atleast_1d(outputs[var].offset)[0] == 1 and np.atleast_1d(outputs[var].offset)[1] == 1)):
                        tmp = outputs[var].dense()
                        if not np.any(outputs[var].isnan(tmp)):
                            outputs[var] = tmp

        # substitute collected values
        for var in outputs:
            i = -1
            for k, kid in enumerate(self.keys()):
                if kid.lower().startswith(var.lower() + '('):
                    del self[kid]
                    i = k

            self.insert(i, var, outputs[var])

        self._collect = True

    def collect(self,value):
        self._collect=value
        for kid in list(self.keys()):
            if isinstance(self[kid],NamelistName):
                self[kid].collect(value)

    def __setitem__(self, key, value):
        if isinstance(value,dict) and not isinstance(value,NamelistName):
            tmp=NamelistName(index_offset=self.index_offset)
            tmp.update(value)
            value=tmp
        return super(NamelistName,self).__setitem__(key,value)

    def __getitem__(self, key):
        key=self._keyCaseInsensitive(key)

        #handle internal key indexing
        if key not in list(self.keys()) and re.match(r'.*\(.*\)',key):
            index=np.array(list(map(int,re.sub(r'(.*)\((.*)\)',r'\2',key).split(','))))
            key=re.sub(r'(.*)\((.*)\)',r'\1',key)
        else:
            index=None

        try:
            tmp=super(NamelistName, self).__getitem__(key)
        except KeyError:
            raise

        #return (indexed) object
        if index is None or not len(np.shape(tmp)):
            return tmp
        else:
            offset=-1
            while index[-1]==1 and len(index)>len(np.atleast_1d(tmp).shape):
                index=index[:-1]
            index+=offset
            return tmp[tuple(index)]

    def __setstate__(self,tmp):
        self.__dict__.update(tmp[0])
        if hasattr(self,'_collect'):
            _collectBCKP=self._collect
            self.collect(False)
        for key,value in zip(self.keyOrder,tmp[1]):
            self[key]=self._setLocation(key,value)
        if hasattr(self,'_collect'):
            self.collect(_collectBCKP)

    def __deepcopy__(self,memo={}):
        if hasattr(self,'_collect'):
            _collectBCKP=self._collect
        else:
            _collectBCKP=False
        self.collect(False)
        tmp=pickle.loads(pickle.dumps(self,pickle.HIGHEST_PROTOCOL))
        tmp.collect(_collectBCKP)
        self.collect(_collectBCKP)
        return tmp

class NamelistFile(NamelistName):
    """
    FORTRAN namelist file object, which can contain multiple namelists blocks

    :param filename: filename to be parsed

    :param input_string: input string to be parsed (preceeds filename)

    :param nospaceIsComment: whether a line which starts without a space should be retained as a comment. If None, a "smart" guess is attempted

    :param outsideOfNamelistIsComment: whether the content outside of the namelist blocks should be retained as comments. If None, a "smart" guess is attempted

    :param retain_comments: whether comments should be retained or discarded

    :param skip_to_symbol: string to jump to for the parsing. Content before this string is ignored

    :param collect_arrays: whether arrays defined throughout the namelist should be collected into single entries
     (e.g. a=5,a(1,4)=0)

    :param multiDepth: wether nested namelists are allowed

    :param bang_comment_symbol: string containing the characters that should be interpreted as comment delimiters.

    :param equals: how the equal sign should be written when saving the namelist

    :param compress_arrays: compress repeated elements in an array by using v*n namelist syntax

    :param max_array_chars: wrap long array lines

    :param explicit_arrays: (True,False,1) whether to place `name(1)` in front of arrays.
                            If `1` then (1) is only placed in front of arrays that have only one value.

    :param separator_arrays: characters to use between array elements

    :param split_arrays: write each array element explicitly on a separate line
                         Specifically this functionality was introduced to split TRANSP arrays

    :param idlInput: whether to interpret the namelist as IDL code
    """
    def __init__(self, filename=None, input_string='', nospaceIsComment=None,
                 outsideOfNamelistIsComment=None, retain_comments=True,
                 skip_to_symbol=None, collect_arrays=True, multiDepth=True,
                 bang_comment_symbol='!;', equals=' = ',
                 compress_arrays=True, max_array_chars=79, explicit_arrays=False,
                 separator_arrays=' ', split_arrays=False,
                 idlInput=False, end_token='/', index_offset=False, **kw):

        NamelistName.__init__(self,index_offset=index_offset)

        self.filename=filename
        self.retain_comments=retain_comments
        self.nospaceIsComment=nospaceIsComment
        self.outsideOfNamelistIsComment=outsideOfNamelistIsComment
        self.skip_to_symbol = skip_to_symbol
        self.collect_arrays = collect_arrays
        self.multiDepth = multiDepth
        self.idlInput = bool(idlInput)
        self.bang_comment_symbol = bang_comment_symbol
        self._equals=equals
        self.compress_arrays=compress_arrays
        self.max_array_chars=max_array_chars
        self.explicit_arrays=explicit_arrays
        self.separator_arrays=separator_arrays
        self.split_arrays=split_arrays
        self.end_token=end_token

        if idlInput:
            self.bang_comment_symbol=';'
            self.nospaceIsComment=False
            self.outsideOfNamelistIsComment=False
            self.compress_arrays=False
            self.max_array_chars=None
            self.separator_arrays=','

        self.dynaLoad=False
        if input_string!='':
            if self.filename is not None:
                with open(self.filename,'w') as f:
                    f.write(input_string)
            else:
                self.parse( input_string.splitlines(True) )
        if self.filename is not None:
            self.dynaLoad=True

    def parse(self, content):
        self.clear()
        if not len(content):
            return

        if _debug():
            start = time.time()

        # ignore cvs version line by adding a bang comment
        for k,line in enumerate(content):
            if '$Id' in line:
                content[k] = re.sub(r'.*\$Id([^$]*?)\$',self.bang_comment_symbol[0]+r' $Id\1$',line)

        # nospace should be interpreted as a comment if I do not find any line which starts by a letter and has a variable assignment
        if self.nospaceIsComment is None:
            self.nospaceIsComment=True
            for line in content:
                line=line.strip('\n')
                for k in self.bang_comment_symbol: #skip commented lines
                    line=line.split(k)[0]
                line=line.rstrip()
                if not len(line):  # Skip lines without length
                    continue
                elif re.search(r'^\s', line):  # Skip lines with leading space
                    continue
                elif re.search(r'(?i)^[&$][_a-z0-9]', line):  # Begin/end of namelist on first char
                    self.nospaceIsComment=False
                elif re.search(r'^[&/$]', line):  # End of namelist on first char
                    self.nospaceIsComment = False
                elif re.search(r'(?i)^[_a-z0-9]\w*\s*=\s*[-\'\"\.]?\w*', line):  # Variable assignment on first char
                    self.nospaceIsComment = False
                    # Variable assignment on first char
                elif re.search(r'(?i)^[_a-z0-9](\([,0-9]\))*\w*\s*=\s*[-\'\"\.]?\w*', line):
                    self.nospaceIsComment = False

        # see if there are variable assignements outside of namelist
        if self.outsideOfNamelistIsComment is None:
            self.outsideOfNamelistIsComment=True
            inside=False
            for line in content:
                line=line.strip('\n')
                for k in self.bang_comment_symbol: #skip commented lines
                    line=line.split(k)[0]
                line=line.rstrip()
                if not len(line): #skip lines without length
                    continue
                elif re.search('(?i)^[_a-z0-9]', line) and self.nospaceIsComment:  # Skip if it's a line commment
                    continue
                elif re.search(r'(?i)^\s*[&$][_a-z0-9]', line):  # Begin of namelist --> inside=True
                    inside = True
                elif re.search(r"(?i)^\s*[&/$]", line):  # End of namelist --> inside=False
                    inside = False
                elif not inside:
                    tmp1 = re.findall(r'(?i)^[_a-z0-9](\([,0-9]\))*\w*\s*=\s*[-\'\"\.]?\w*', line)
                    if len(tmp1) == 1:  # Only one variable per line
                        tmp2 = re.findall(r'(?i)^[_a-z0-9](\([,0-9]\))*\w*\s*=\s*[-\'\"\.]?\w*', line)
                        if len(tmp2) and tmp2[0] == tmp1[0]:  # There is no text before the first variable
                            self.outsideOfNamelistIsComment = False

        if _debug():
            for k in ['nospaceIsComment','outsideOfNamelistIsComment']:
                print(k,getattr(self,k))
            print('read in '+format(time.time() - start,'3.3f')+ ' sec.')

        rules={}
        rules['single']=0
        rules['double']=0
        rules['round_bracket']=0
        rules['bang_comment']=0
        rules['line_comment']=0
        if self.skip_to_symbol is None:
            tmpj=''.join(content)
        else:
            for si,s in enumerate(content):
                if self.skip_to_symbol in s:
                   break
            if si==len(content):
                si=0
            tmpj = ''.join(content[si:])
        item_offset=0
        item_length=0
        item_prepend=''
        tmp=[]

        if self.idlInput:
            # in IDL language `&` is like a new line
            # here it's safer to convert to a space
            tmpj=re.sub('&',' ',tmpj)

        inside=0
        tmpj = re.sub(r"(?i)[&/$]end(_\S*)?","$",tmpj).rstrip()+'\n'*3

        kk=0
        for cm,c,cp in zip(*['\n'+tmpj,tmpj,tmpj[1:]+'\n']):
            kk+=1
            if _debug()>1:
                print(cm,c,cp,tmp)

            if cm=='\\':
                #if it's an escaped character, then it's nothing special, by definition
                item_length+=1
                continue

            if c=='\n':
                #if it's a new line, then it's a new record and all rules should be reset
                for key in list(rules.keys()):
                    rules[key]=0

            #handle comments outside of namelist
            if self.outsideOfNamelistIsComment:

                #bang comments should always be considered comments
                if c in self.bang_comment_symbol:
                    rules['bang_comment']=1

                #when namelist block starts or the file ends, assign all to be comment
                if not rules['bang_comment'] and re.match(r'(?i)[$&][_a-z0-9]',c+cp) or kk==len(tmpj):
                    if not inside:
                        item_=tmpj[item_offset:item_offset+item_length]
                        item_=re.sub('[\n \t]','',item_)
                        if len(item_): #do not store empty blocks
                            if not len(tmp):
                                item=u'\1'+tmpj[item_offset:item_offset+item_length].rstrip(' \t')[:-1]
                            else:
                                item=u'\1'+tmpj[item_offset:item_offset+item_length].strip(' \t')[1:-1]
                            item_offset+=item_length
                            item_length=0
                            item_prepend=''
                            tmp.append(item)

                #check if a namelist block is starting
                if not rules['bang_comment'] and re.match(r'(?i)[$&][_a-z0-9]',c+cp):
                    inside+=1
                    if _debug()>1:
                        print('namelist block starts')

            #if inside of a namelist block or if variables outside namelist blocks are not comments
            if inside or not self.outsideOfNamelistIsComment:

                #this is a new item
                if (c in ' \t,=' and not np.any([rules[key] for key in list(rules.keys())])) or c=='\n':
                    if item_length>0:
                        item=item_prepend+tmpj[item_offset:item_offset+item_length]
                        item_offset+=item_length
                        item_length=0
                        item_prepend=''
                        if len(item.strip('\n \t,=')):
                            tmp.append(item.strip('\n \t,='))
                        if c=='=':
                            tmp.append(c)

                #this is the continuation of the current item
                else:
                    if c=='(' and not np.any([rules[key] for key in list(rules.keys()) if key!='round_bracket']):
                        rules['round_bracket']+=1
                    elif c==')' and not np.any([rules[key] for key in list(rules.keys()) if key!='round_bracket']):
                        rules['round_bracket']-=1
                    elif c=="'" and not np.any([rules[key] for key in list(rules.keys()) if key!='single']):
                        rules['single']=(rules['single']+1)%2
                    elif c=='"' and not np.any([rules[key] for key in list(rules.keys()) if key!='double']):
                        rules['double']=(rules['double']+1)%2
                    elif (c in self.bang_comment_symbol) and not np.any([rules[key] for key in list(rules.keys()) if key!='bang_comment']):
                        rules['bang_comment']=1
                    elif (self.nospaceIsComment and not self.outsideOfNamelistIsComment) and (cm=='\n' and c not in ' \t') and not np.any([rules[key] for key in list(rules.keys()) if key!='line_comment']):
                        rules['line_comment']=1
                        item_prepend=u'\1'

                #when namelist block ends
                if self.outsideOfNamelistIsComment:
                    if inside and not np.any([rules[key] for key in list(rules.keys())]) and ((cm=='/' and c=='\n')):
                        inside-=1
                        # if namelist block ends and last item had / then
                        # reasonable to assume that / was the termination
                        if inside==0 and len(tmp[-1])>1 and tmp[-1][-1]=='/':
                            tmp[-1]=tmp[-1][:-1]
                        if _debug()>1:
                            print('namelist block ends')

            item_length=item_length+1

        if _debug():
            print('b in '+format(time.time() - start,'3.3f')+ ' sec.')
        tmp = [re.sub(r"(?i)[&/$]end","/",line) for line in tmp]
        if _debug()>1:
            print(tmp)

        def uniqueSortedComment(itm,tree):
            out='__comment'+omfit_hash(str(itm),10)+'__'
            while out in tree:
                out='__comment'+omfit_hash(str(out),10)+'__'
            return out

        #build the namelist tree
        tree=self
        vlist=None
        for k,item in enumerate(tmp):
            if re.match(r'^[&$].+',item):
                if not self.multiDepth and '__parent__' in tree:
                    tmptree=tree['__parent__']
                    del tree['__parent__']
                    tree=tmptree
                    vlist=None
                #is it a namelist header
                nlist=re.sub(r'^[&$](.*)',r'\1',item)
                if nlist not in tree:
                    tree[nlist]=NamelistName(index_offset=self.index_offset)
                tree[nlist]['__parent__']=tree
                tree=tree[nlist]
            elif item[0]==u'\1':
                #it's a line comment
                if self.retain_comments:
                    tree[uniqueSortedComment(item[1:],tree)]=item[1:]+'\n'
            elif item[0] in self.bang_comment_symbol:
                #it's a bang comment
                if self.retain_comments:
                    tree[uniqueSortedComment(item,tree)]=item+'\n'
            elif item in ['/','$','&']:
                #is it a namelist close
                try:
                    tmptree=tree['__parent__']
                    del tree['__parent__']
                    tree=tmptree
                    vlist=None
                except Exception:
                    if self.multiDepth:
                        warnings.warn('Some problem with namelist closure near: +\n\n'+' '.join(tmp[k-100:k]))
            elif k!=len(tmp)-1 and tmp[k+1]=='=':
                #If the next item is an equal sign, this means that it's a variable name
                tmp[k]=item=re.sub(' ','',item) #variable names in namelists cannot have spaces
                tree[item]=[]
                vlist=tree[item]
            elif item=='=':
                pass
            else:
                if not(vlist is None):
                    if '*' in item and re.match(r'([0-9]+)\*(.+)',item):
                        nrep=int(re.sub(r'([0-9]+)\*(.+)',r'\1',item))
                        val=re.sub(r'([0-9]+)\*(.+)',r'\2',item)
                        vlist.extend([val]*nrep)
                    elif self.idlInput:
                        vlist.append(item.strip('[]'))
                    else:
                        vlist.append(item)
        if _debug():
            print('c in '+format(time.time() - start,'3.3f')+ ' sec.')

        if not self.multiDepth:
            fun_str = lambda orig:interpreter(orig,escaped_strings='fortran')
        else:
            fun_str = lambda orig:interpreter(orig,escaped_strings=True)

        #now interpret the list of stings
        def i_traverse(me):
            if '__parent__' in list(me.keys()):
                del me['__parent__']
            for kid in list(me.keys()):
                if isinstance(me[kid],dict):
                    i_traverse(me[kid])
                elif not re.match(comment_ptrn,kid):
                    fun=fun_str
                    if len(me[kid])>1:
                        item0 = copy.deepcopy(me[kid][0])
                        try:
                            int(item0)
                            fun = int
                        except Exception:
                            try:
                                float(item0)
                                fun = float
                            except Exception:
                                pass
                    try:
                        me[kid] = list(map(fun, me[kid]))
                    except Exception:
                        me[kid] = list(map(fun_str, me[kid]))
                    if isinstance(me[kid],str):
                        pass
                    elif len(me[kid])>1:
                        me[kid]=np.array(me[kid])
                    elif len(me[kid])==1:
                        me[kid]=me[kid][0]
                    else:
                        me[kid]=None
        i_traverse(self)
        if _debug():
            print('d in '+format(time.time() - start,'3.3f')+ ' sec.')

        #collect arrays
        if self.collect_arrays:
            try:
                self.collectArrays(self.collect_arrays)
            except Exception as _excp:
                printw('Could not collect arrays: '+repr(_excp))

        if _debug():
            print('end of loading namelist')
            print('Parsing namelist in '+format(time.time() - start,'3.3f')+ ' sec.')

    def saveas(self, filename):
        """save namelist to a new file"""
        self.filename=os.path.abspath(filename)
        directory=os.path.split(os.path.abspath(filename))[0]
        if os.path.exists(directory)==0:
            os.makedirs(directory)
        self.save()

    @dynaSave
    def save(self, fp=None):
        if self.filename == None and fp == None:
            return
        #this whole pointer trickery is to allow saving of namelists in EFIT g-file
        if fp is not None:
            f=fp
        else:
            f=open(self.filename, 'w+')

        def f_traverse(me,level):
            for kid in list(me.keys()):
                if _debug():
                    print('Writing ',kid)
                if isinstance_str(me[kid], ['OMFITexpression', 'OMFITiterableExpression']):
                    meKid=me[kid]._value_()
                else:
                    meKid=me[kid]

                if isinstance(meKid,dict):
                    level+=1
                    f.writelines(' '*level+'&'+kid+'\n')
                    f_traverse(meKid,level)
                    f.writelines(' '*level+self.end_token+'\n')
                    level-=1
                elif re.match(comment_ptrn,kid):
                    f.writelines(meKid.rstrip()+'\n')
                else:
                    #if it is an array
                    if (isinstance(meKid,np.ndarray) or
                        isinstance(meKid,sparray) or
                        isinstance(meKid,list) or
                        re.search(r'^\[(([ \n]*[0-9\-+.ji]+)*)\]$',str(meKid))):
                        #handle split_arrays functionality by looping over individual segments
                        if not self.split_arrays:
                            kid0,meKid0=[kid],[meKid],
                        elif callable(self.split_arrays): #this cannot be saved, but it is useful for development of the _split_arrays_function
                            kid0,meKid0=self.split_arrays(kid,meKid)
                        else:
                            kid0,meKid0=_split_arrays_function(kid,meKid)
                        for kid,meKid in zip(kid0,meKid0):
                            if isinstance(meKid,sparray):
                                tmp=meKid.fortran_repr()
                                for item,value in zip(list(tmp.keys()),list(tmp.values())):
                                    line=' '*level+kid+item+self._equals
                                    if self.idlInput: line+='['
                                    tmp=array_encoder(value,line,level,separator_arrays=self.separator_arrays,compress_arrays=self.compress_arrays,max_array_chars=self.max_array_chars,escaped_strings=['fortran',True][self.multiDepth])
                                    if self.idlInput: tmp+=']'
                                    f.writelines(tmp.rstrip()+'\n')
                            elif isinstance(meKid,np.ndarray) and len(meKid.shape)>1:
                                kidShape=list(meKid.shape)
                                kidShape[0]=1
                                i=0
                                fi=lambda i=i:'0:'+str(i)
                                ndim=np.reshape(eval("np.mgrid["+','.join([fi(i) for i in kidShape])+"]"),(len(kidShape),-1))
                                tmpData=np.reshape(meKid,(meKid.shape[0],-1))
                                for kk,k in enumerate(range(len(ndim[0]))):
                                    line=' '*level+kid+'('+','.join([str(i+1) for i in ndim[:,k]])+')'+self._equals
                                    if self.idlInput: line+='['
                                    tmp=array_encoder(tmpData[:,kk],line,level,separator_arrays=self.separator_arrays,compress_arrays=self.compress_arrays,max_array_chars=self.max_array_chars,escaped_strings=['fortran',True][self.multiDepth])
                                    if self.idlInput: tmp+=']'
                                    f.writelines(tmp.rstrip()+'\n')
                            else:
                                if re.match(r'.*\(\)',kid):
                                    kid=kid[:-2]
                                elif self.explicit_arrays is 1 and len(meKid)==1 and not re.match(r'.*\(.*\)',kid):
                                    kid=kid+'(1)'
                                elif self.explicit_arrays is True and not re.match(r'.*\(.*\)',kid):
                                    kid=kid+'(1)'
                                line=' '*level+kid+self._equals
                                if self.idlInput: line+='['
                                tmp=array_encoder(meKid,line,level,separator_arrays=self.separator_arrays,compress_arrays=self.compress_arrays,max_array_chars=self.max_array_chars,escaped_strings=['fortran',True][self.multiDepth])
                                if self.idlInput: tmp+=']'
                                f.writelines(tmp.rstrip()+'\n')
                    else:
                        f.writelines((' '*level+kid+self._equals+encoder(meKid,escaped_strings=['fortran',True][self.multiDepth])).rstrip()+'\n')

        f_traverse(self,0)
        if fp is None:
            f.close()

    @dynaLoad
    def load(self, filename=''):
        """Load namelist from file."""
        self.clear()
        self.dynaLoad=False
        if filename!='':
            self.filename=filename
        with open(self.filename,'r') as f:
            self.parse( f.readlines() )
        return self

class fortran_environment(object):
    '''
    Environment class used to allow FORTRAN index convention of sparrays in a namelist
    '''

    def __init__(self, nml):
        self.nml = nml
        self.sp = traverse(self.nml, onlyLeaf=(sparray,))

    def __enter__(self):
        self.index_offset_bkp = {}
        for item in self.sp:
            self.index_offset_bkp[item] = getattr(eval("self.nml" + item),'index_offset',0)
            eval("self.nml" + item).index_offset = True

    def __exit__(self, type, value, traceback):
        for item in self.sp:
            eval("self.nml" + item).index_offset = self.index_offset_bkp[item]

class sparray(object):
    """
    Class for n-dimensional sparse array objects using Python's dictionary structure.
    based upon: http://www.janeriksolem.net/2010/02/sparray-sparse-n-dimensional-arrays-in.html
    """

    def __init__(self, shape=None, default=np.nan, dtype=None, wrap_dim=0, offset=0, index_offset=False):
        if dtype is None:
            self.dtype = None
        elif callable(dtype):
            self.dtype = dtype = np.array(dtype(0)).dtype
        else:
            self.dtype = dtype = np.array(dtype).dtype
        self.default = np.array(default)
        if self.dtype is not None:
           self.default = np.array(default).astype(dtype)
        elif dtype is None:
            self.dtype = dtype = self.default.dtype
        if shape is None:
            self.shape = ()
        else:
            self.shape = tuple(shape)
        self.ndim = len(self.shape)
        self.wrap_dim = wrap_dim
        self.__data = self.data = OrderedDict()
        self.offset = np.atleast_1d(offset)
        self.index_offset = index_offset

    def _index_offset(self, index, mult=-1):
        if not self.index_offset:
            return index
        index = np.atleast_1d(index)
        _index = []
        offset = np.zeros(index.shape) + self.offset
        for i, o in zip(index, offset):
            if isinstance(i, slice):
                start, stop, step = i.start, i.stop, i.step
                if start is not None:
                    start = int(start + o * mult)
                if stop is not None:
                    stop = int(stop + o * mult)
                _index.append(slice(start, stop, step))
            else:
                _index.append(int(i + o * mult))
        return tuple(_index)

    def _setitem_base(self, index, value):
        """ set value to position given in index, where index is a tuple. """
        # adjust number of dimensions
        index=tuple(map(int,index))
        if len(index) > len(self.shape):
            tmp = list(np.array(index) + 1)
            tmp[:len(self.shape)] = self.shape
            self.shape = tuple(tmp)
        # adjust dimensions
        for k, dim in enumerate(np.array(index) + 1):
            if dim > self.shape[k]:
                tmp = list(self.shape)
                tmp[k] = dim
                self.shape = tuple(tmp)
        self.ndim = len(self.shape)
        if self.dtype is None:
            self.dtype = np.array(value).dtype
            self.default = np.array(self.default).astype(self.dtype)
        self.__data[index] = np.atleast_1d([value]).astype(self.dtype)

    def __setitem__(self, index, value):
        """ set value to position given in index, where index is a tuple. """
        index = self._index_offset(index, -1)
        if isinstance(value, np.ndarray):
            _index = list(index)
            for k in range(len(value)):
                self._setitem_base(tuple(_index), value[k])
                _index[self.wrap_dim] += 1
        else:
            self._setitem_base(index, value)

    def __getitem__(self, index):
        """ get value at position given in index, where index is a tuple. """
        index = self._index_offset(index, -1)
        try:
            return self.dense()[index]
        except IndexError:
            return index*0+self.default

    def __delitem__(self, index):
        """ index is tuples of element to be deleted. """
        index = self._index_offset(index)
        if index in self.__data:
            del (self.__data[index])

    def __binary_op__(self, other, op):
        if isinstance(other,sparray):
            if self.shape == other.shape:
                tmp = copy.deepcopy(self)
                for k in set.difference(set(tmp.__data.keys()), set(other.__data.keys())):
                    tmp.__data[k] = eval('tmp.__data[k].' +op+ '(other.default)')
                tmp.default = eval('tmp.default.' + op+ '(other.default)')
                for k in list(other.__data.keys()):
                    old_val = tmp.__data.setdefault(k, tmp.default)
                    tmp.__data[k] = eval('old_val.' +op+ '(other.__data[k])')
                return tmp
            else:
                raise ValueError('Array sizes do not match. ' + str(self.shape) + ' versus ' + str(other.shape))
        else:
            return eval('self.dense().'+op+'(other)')

    def __unary_op__(self, op):
        tmp = copy.deepcopy(self)
        for k in tmp.__data.keys():
            tmp.__data[k] = eval('tmp.__data[k].' +op+ '()')
        tmp.default = eval('tmp.default.' + op+ '()')
        return tmp

    def __str__(self):
        return str(self.dense())

    def dense(self):
        """Convert to dense NumPy array"""
        out = (self.default * np.ones(self.shape)).astype(self.dtype)
        for ind in self.__data:
            out[ind] = self.__data[ind].astype(self.dtype)
        return out

    def sum(self):
        """Sum of elements"""
        s = self.default * np.array(self.shape).prod()
        for ind in self.__data:
            s += (self.__data[ind] - self.default)
        return s

    def fortran(self, index, value):
        # Drop leading zeros otherwise the index is evaluated as an octal.
        index = delete_leading_zeros(index)
        index = eval('(' + re.sub('([0-9]+):', r'\1', index.split('(')[-1].strip(')')) + ')')
        self.__setitem__(index, value)
        return self

    def isnan(self,x):
        if np.iterable(x):
            x=np.array(x)
            i=np.where(x==-9223372036854775808)
            x=x.astype(float)
            x[i]=np.nan
            return np.isnan(np.array(x))
        elif np.isnan(x):
            return True
        elif x==np.array(np.nan).astype(np.array(x).dtype):
            return x==np.array(np.nan).astype(np.array(x).dtype)

    def fortran_repr(self):
        # print strips
        strips = OrderedDict()
        for item in list(self.__data.keys()):
            item = tuple(item)
            strips.setdefault(item[1:], item)
            strips[item[1:]] = tuple(map(int, np.min([strips[item[1:]], item], 0)))

        # prints
        outputs = OrderedDict()
        if not self.isnan(self.default):
            outputs['']=self.default

        for item in list(strips.values()):
            tmp=list(item)
            tmp[0]=slice(None)
            tmp=tuple(tmp)

            # do not use self[tmp] to avoid self._index_offset
            dense_tmp=self.dense()[tmp]

            if self.isnan(self.default):
                ix_=np.where(self.isnan(dense_tmp))[0]
            else:
                ix_=np.where(dense_tmp==self.default)[0]

            #split where default value is found
            substrips_=np.split(dense_tmp,ix_)
            substrips=[]
            ix=[]
            ix_=[0]+list(ix_)
            for ki,ks in zip(ix_,substrips_):
                if len(ks)>1:
                    ix.append(ki)
                    ix.append(ki+1)
                    substrips.append(ks[:1])
                    substrips.append(ks[1:])
                elif len(ks):
                    ix.append(ki)
                    substrips.append(ks)

            #print substrips
            #note: values that are equal to the default value or that are NaN will not be printed
            item=item+self.offset
            for k,v in enumerate(substrips):
                if not len(v):
                    continue
                elif self.isnan(v[0]):
                    continue
                elif v[0]==self.default:
                    continue
                item[0]=ix[k]+self.offset[0]
                if len(item)==1:
                    outputs['(%d)'%item.astype(int)] = v.astype(self.dtype)
                else:
                    outputs[str(re.sub(' ', '', repr(tuple(item.astype(int)))))] = v.astype(self.dtype)

        return outputs

    def __str__(self):
        tmp = self.fortran_repr()
        out = []
        for item, value in zip(list(tmp.keys()), list(tmp.values())):
            out.append('%s: %s' % (str(item)[1:-1], str(value)))
        return '\n'.join(out)

    def __repr__(self):
        return self.__str__()

    def __tree_repr__(self):
        return self.dense(), []

    def __iter__(self):
        for i in self.dense():
            yield i

    # binary operations
    def lt(self, other):
        return self.__binary_op__(other, 'lt')

    def le(self, other):
        return self.__binary_op__(other, 'le')

    def eq(self, other):
        return self.__binary_op__(other, 'eq')

    def ge(self, other):
        return self.__binary_op__(other, 'ge')

    def gt(self, other):
        return self.__binary_op__(other, 'gt')

    def __lt__(self, other):
        return self.__binary_op__(other, '__lt__')

    def __le__(self, other):
        return self.__binary_op__(other, '__le__')

    def __eq__(self, other):
        return self.__binary_op__(other, '__eq__')

    def __ne__(self, other):
        return self.__binary_op__(other, '__ne__')

    def __ge__(self, other):
        return self.__binary_op__(other, '__ge__')

    def __gt__(self, other):
        return self.__binary_op__(other, '__gt__')

    def is_(self, other):
        return self.__binary_op__(other, 'is_')

    def is_not(self, other):
        return self.__binary_op__(other, 'is_not')

    def add(self, other):
        return self.__binary_op__(other, 'add')

    def __add__(self, other):
        return self.__binary_op__(other, '__add__')

    def __radd__(self, other):
        return self.__binary_op__(other, '__radd__')

    def and_(self, other):
        return self.__binary_op__(other, 'and_')

    def __and__(self, other):
        return self.__binary_op__(other, '__and__')

    def __rand__(self, other):
        return self.__binary_op__(other, '__rand__')

    def floordiv(self, other):
        return self.__binary_op__(other, 'floordiv')

    def __floordiv__(self, other):
        return self.__binary_op__(other, '__floordiv__')

    def __rfloordiv__(self, other):
        return self.__binary_op__(other, '__rfloordiv__')

    def index(self, other):
        return self.__binary_op__(other, 'index')

    def __index__(self, other):
        return self.__binary_op__(other, '__index__')

    def lshift(self, other):
        return self.__binary_op__(other, 'lshift')

    def __lshift__(self, other):
        return self.__binary_op__(other, '__lshift__')

    def __rlshift__(self, other):
        return self.__binary_op__(other, '__rlshift__')

    def mod(self, other):
        return self.__binary_op__(other, 'mod')

    def __mod__(self, other):
        return self.__binary_op__(other, '__mod__')

    def __rmod__(self, other):
        return self.__binary_op__(other, '__rmod__')

    def mul(self, other):
        return self.__binary_op__(other, 'mul')

    def __mul__(self, other):
        return self.__binary_op__(other, '__mul__')

    def __rmul__(self, other):
        return self.__binary_op__(other, '__rmul__')

    def matmul(self, other):
        return self.__binary_op__(other, 'matmul')

    def __matmul__(self, other):
        return self.__binary_op__(other, '__matmul__')

    def or_(self, other):
        return self.__binary_op__(other, 'or_')

    def __or__(self, other):
        return self.__binary_op__(other, '__or__')

    def __ror__(self, other):
        return self.__binary_op__(other, '__ror__')

    def pow(self, other):
        return self.__binary_op__(other, 'pow')

    def __pow__(self, other):
        return self.__binary_op__(other, '__pow__')

    def __rpow__(self, other):
        return self.__binary_op__(other, '__rpow__')

    def rshift(self, other):
        return self.__binary_op__(other, 'rshift')

    def __rshift__(self, other):
        return self.__binary_op__(other, '__rshift__')

    def __rrshift__(self, other):
        return self.__binary_op__(other, '__rrshift__')

    def sub(self, other):
        return self.__binary_op__(other, 'sub')

    def __sub__(self, other):
        return self.__binary_op__(other, '__sub__')

    def __rsub__(self, other):
        return self.__binary_op__(other, '__rsub__')

    def truediv(self, other):
        return self.__binary_op__(other, 'truediv')

    def __truediv__(self, other):
        return self.__binary_op__(other, '__truediv__')

    def __rtruediv__(self, other):
        return self.__binary_op__(other, '__rtruediv__')

    def __divmod__(self, other):
        return self.__binary_op__(other, '__divmod__')

    def __rdivmod__(self, other):
        return self.__binary_op__(other, '__rdivmod__')

    def xor(self, other):
        return self.__binary_op__(other, 'xor')

    def __xor__(self, other):
        return self.__binary_op__(other, '__xor__')

    def __rxor__(self, other):
        return self.__binary_op__(other, '__rxor__')

    # unary operations
    def bool(self):
        return self.__unary_op__('bool')

    def __bool__(self):
        return self.__unary_op__('__bool__')

    def __nonzero__(self):
        return self.__unary_op__('__nonzero__')

    def real(self):
        return self.__unary_op__('real')

    def imag(self):
        return self.__unary_op__('imag')

    def not_(self):
        return self.__unary_op__('not_')

    def __not__(self):
        return self.__unary_op__('__not__')

    def truth(self):
        return self.__unary_op__('truth')

    def abs(self):
        return self.__unary_op__('abs')

    def __abs__(self):
        return self.__unary_op__('__abs__')

    def inv(self):
        return self.__unary_op__('inv')

    def invert(self):
        return self.__unary_op__('invert')

    def __inv__(self):
        return self.__unary_op__('__inv__')

    def __invert__(self):
        return self.__unary_op__('__invert__')

    def neg(self):
        return self.__unary_op__('neg')

    def __neg__(self):
        return self.__unary_op__('__neg__')

    def pos(self):
        return self.__unary_op__('pos')

    def __pos__(self):
        return self.__unary_op__('__pos__')

    def __float__(self):
        return self.__unary_op__('__float__')

    def __complex__(self):
        return self.__unary_op__('__complex__')

    def __oct__(self):
        return self.__unary_op__('__oct__')

    def __hex__(self):
        return self.__unary_op__('__hex__')

    def __trunc__(self):
        return self.__unary_op__('__trunc__')

    __hash__ = None

def delete_leading_zeros(number_str):
    '''
    Delete leading zeros from array indices

    :param number_str: A string containing comma separated dimension indices and colon separated ranges and strides, such as '008:010,0001:0002'

    :return: String with leading zeros stripped, such as '8:10,1:2'
    '''
    return ','.join([':'.join([colon_split.lstrip('0') if colon_split.lstrip('0') else '0'
                                  for colon_split in comma_split.split(':')])
                                        for comma_split in number_str.split(',')])

#--------------------------------
#--------------------------------
#--------------------------------
############################################
if __name__ == '__main__':
    OMFITsrc = os.path.abspath(os.path.dirname(__file__) + os.sep + '..')
    test_classes_main_header()

    ranged_nl = NamelistFile(input_string='''
                             &test
                                 entry(1,1) = 1
                                 entry(2,1:2) = 1, 2
                                 entry(3,1) = 1
                                 entry(3,2) = 2
                                 entry(3,3) = 3
                                 entry(4,1) = 4
                                 entry(5,1:2) = 1, 2
                                 entry(6:7,1) = 6, 7
                             /
                             ''')
    assert ranged_nl['test']['entry'][1,1]==2 # Note that internal slicing is python-style based on 0
    ranged_nl['test']['entry'][1,1]=1
    assert ranged_nl['test']['entry'][1,1]==1
    assert ranged_nl['test']['entry'][4,1]==2

    ranged_nl.filename = '/tmp/ranged_nl.dat'
    ranged_nl.save()
    ranged_nl2 = NamelistFile('/tmp/ranged_nl.dat')
    ranged_nl2.load()
    failed_nl = NamelistFile(input_string='''
                             &test
                                entry(1:2,1:2) = 1
                             /
                             ''')
    failed_nl.load()
    foo = NamelistFile(OMFITsrc+'/../samples/k139817.00109')
    foo.load()
    print('---')
    foo['asd(1,1)']=np.array([1,2,3])
    foo['asd(1,2)']=np.array([4,5,6])
    foo.collectArrays()

    A2_Z=list(range(2))
    ILA_Z=list(range(3))
    antennas=list(range(6))
    foo['YGEOANT_A'] = YGEOANT_A = sparray( shape=(max(len(A2_Z),len(ILA_Z)),6),default=0)
    for a, ant in enumerate(antennas):
            for i in range( len(A2_Z) ):
                YGEOANT_A[ ( i+1, a+1 ) ] = (i+1)*10+a+1

    foo.filename = '/tmp/%s/%s'%(os.environ.get('USER','nobody'),os.path.split(foo.filename)[1])
    foo.save()
    foo=NamelistFile(foo.filename)

    with open(foo.filename) as f:
        print(f.read())
    #print(foo)

    #asd=copy.deepcopy(foo)
    #foo.saveas('../samples/inone.03680_saved')
    #for k in open('../samples/inone.03680_saved','r').readlines():
    #    print(k.strip('\n'))

#     test='''
# because of this line it is assumed that everything outside the namelist blocks is a comment
# !b=1
# ac=4
# &my_namelist
#     hello='1 \$ 1'
#     array=1 2 3 3 4 5 6 6 6 7 8
#     array2=1 2 2*3 4 5 3*6 7 8
#     array3=asd asd asd
#     array4=3*asd
#     array5=
#     aaa(1,1)=4
#     aaa(1, 1)=5
# $END
#  ;this is another comment
# !this is a different comment format
#     '''
#     foo = NamelistFile(input_string=test,bang_comment_symbol='!;')
#     print(foo)
#     print(array_encoder(foo['my_namelist']['array']))
#     print(array_encoder(foo['my_namelist']['array3']))
#
#



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITstrahl']


class OMFITstrahl(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with STRAHL input files

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.readlines()
        cvs = -1
        for line in lines:
            if line.startswith('cv '):
                cvs += 1
                key = cvs
                self[key] = []

            self[key].append(line.strip())
        for key in self:
            name = re.sub(r'cv\s+(.*)', r'\1', self[key][0])
            value = ' '.join([_f for _f in self[key][1:] if _f])
            if not value.startswith("'") and ' ' in value:
                value = np.array(list(map(float, value.split(' '))))
                for item in ['Electron Density', 'Electron Temperature', 'Ion Temperature']:
                    if name.lower().startswith(item.lower()):
                        value = value[1:] * value[0]
                        break
            else:
                value = eval(value)
            self[key] = [name, value]

    def save(self):
        lines = []
        for key, value in list(self.values()):
            lines.append('cv    ' + key)
            if key == 'time-vector':
                for item in value:
                    lines.append('   %f' % item)
            elif isinstance(value, str):
                lines.append('   %s' % repr(value))
            else:
                tmp = value
                for item in ['Electron Density', 'Electron Temperature', 'Ion Temperature']:
                    if key.lower().startswith(item.lower()):
                        tmp = [value[0]] + tolist(value / value[0])
                        break
                lines.append('   %s' % ' '.join(map(str, tolist(tmp))))
            lines.extend(['', ''])
        pprint(lines)
        with open(self.filename, 'w') as f:
            f.write('\n'.join(lines))



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

import xarray
import numpy as np

__all__ = ['OMFITmars', 'OMFITnColumns', 'OMFITmarsProfile']


class OMFITmars(SortedDict, OMFITobject):
    """
    Class used to interface with MARS results directory

    :param filename: directory where the MARS result files are stored

    :param extra_files: Any extra files that should be loaded

    :param Nchi: Poloidal grid resolution for real-space
    """

    def __init__(self, filename=None, extra_files=[], Nchi=None, **kw):

        if Nchi is None:
            Nchi = 200
            printw(f'OMFITmars using default Nchi value of {Nchi}')

        if isinstance(filename, (list, tuple)):
            kw['tunnel'] = filename[2]
            kw['server'] = filename[1]
            filename = filename[0]

        if ',' not in filename:
            outputs = [
                'BPLASMA.OUT',
                'BPLASMA',
                'JACOBIAN.OUT',
                'JACOBIAN',
                'RMZM_F.OUT',
                'RMZM_F',
                'XPLASMA.OUT',
                'XPLASMA',
                'JPLASMA.OUT',
                'JPLASMA',
                'PROFEQ.OUT',
                'PROFEQ',
                'RESULT.OUT',
                'RESULT',
                'FREQUENCIES.OUT',
                'TIMEEVOL.OUT',
                'DWK_ENERGY_DENSITY.OUT',
            ]
            outputs += extra_files
            filename = ','.join([filename + os.sep + x for x in outputs])

        kw['file_type'] = 'dir'
        OMFITobject.__init__(self, filename, **kw)
        self.OMFITproperties.pop('file_type', 'dir')
        self.OMFITproperties['Nchi'] = Nchi
        self.sim = 'sim0'
        SortedDict.__init__(self)
        self.dynaLoad = True

    """
    Reading methods, used to read MARS output files.
    """

    def read_RMZM(self, filename):
        with open(filename) as fin:
            x = [ii.strip() for ii in fin.readlines()]

        RMZM_F = np.zeros((len(x), 4))

        for ii, line in enumerate(x):
            RMZM_F[ii] = line.split()

        # Number of poloidal harmonics for equilibrium (RMZM)
        Nm0 = int(RMZM_F[0, 0])
        Ns1 = int(RMZM_F[0, 1])
        Ns2 = int(RMZM_F[0, 2])

        R0EXP = RMZM_F[0, 3]
        B0EXP = RMZM_F[1, 3]

        Ns = Ns1 + Ns2

        NRATSURF = int(RMZM_F[1, 1])
        Iratsurf = np.array(RMZM_F[2 : NRATSURF + 2, 1], int)
        # s coordinate over full domain
        s = np.array(RMZM_F[1 : Ns + 1, 0], float)
        # s coordinate in plasma
        sp = np.array(RMZM_F[1 : Ns1 + 1, 0], float)
        # s coordinate in vacuum
        svac = np.array(RMZM_F[Ns1 + 1 : Ns + 1, 0], float)
        RM = np.array(RMZM_F[Ns + 1 :, 0] + RMZM_F[Ns + 1 :, 1] * 1j, complex)
        ZM = np.array(RMZM_F[Ns + 1 :, 2] + RMZM_F[Ns + 1 :, 3] * 1j, complex)

        RM = np.reshape(RM, (Nm0, Ns)).T
        ZM = np.reshape(ZM, (Nm0, Ns)).T

        RM[:, 1:] = 2 * RM[:, 1:]
        ZM[:, 1:] = 2 * ZM[:, 1:]
        Nchi = self.OMFITproperties['Nchi']

        self[self.sim]['RM'] = xarray.DataArray(RM, dims=['s', 'm'], coords={'m': range(Nm0), 's': s})
        self[self.sim]['ZM'] = xarray.DataArray(ZM, dims=['s', 'm'], coords={'m': range(Nm0), 's': s})
        self[self.sim]['sp'] = xarray.DataArray(sp, dims=['sp'], coords={'sp': sp}, attrs={'Description': 's-coord in plasma domain'})
        self[self.sim]['svac'] = xarray.DataArray(
            svac, dims=['svac'], coords={'svac': svac}, attrs={'Description': 's-coord in vacuum domain'}
        )
        self[self.sim]['Iratsurf'] = xarray.DataArray(Iratsurf)
        self[self.sim]['R0EXP'] = xarray.DataArray(R0EXP)
        self[self.sim]['B0EXP'] = xarray.DataArray(B0EXP)
        self[self.sim]['Ns1'] = xarray.DataArray(Ns1, attrs={'Description': 'Number of radial points in plasma'})
        self[self.sim]['Ns2'] = xarray.DataArray(Ns2, attrs={'Description': 'Number of radial points in vacuum'})
        self[self.sim]['Nm0'] = xarray.DataArray(Nm0, attrs={'Description': 'Number of poloidal harmonics for equilibrium'})
        self[self.sim]['Nchi'] = xarray.DataArray(Nchi, attrs={'Description': 'Number of points along poloidal angle (from CHEASE)'})

        return RM, ZM

    def read_JPLASMA(self, filename, JNORM=1.0):
        """
        Read JPLASMA.OUT file corresponding to perturbed currents

        :param filename: name of file

        :param JNORM: normalization
        """
        JPLASMA = np.loadtxt(filename)
        # Nm1 = Number of poloidal harmonics for stability (perturbed quantities, M1...M2)
        Nm1, Ns, n, _, _, _ = map(int, JPLASMA[0, :])
        JPLASMA = JPLASMA[1:, :]

        Mm = JPLASMA[:Nm1, 1].astype(int)
        JM1 = JPLASMA[Nm1:, 0] + 1j * JPLASMA[Nm1:, 1]
        JM2 = JPLASMA[Nm1:, 2] + 1j * JPLASMA[Nm1:, 3]
        JM3 = JPLASMA[Nm1:, 4] + 1j * JPLASMA[Nm1:, 5]

        JM1 = np.reshape(JM1, (Nm1, Ns)).T
        JM2 = np.reshape(JM2, (Nm1, Ns)).T
        JM3 = np.reshape(JM3, (Nm1, Ns)).T
        JM1 = JM1 * JNORM
        JM2 = JM2 * JNORM
        JM3 = JM3 * JNORM

        # Note that JM1 is defined at half-points, recompute at integer-points
        x = np.asarray((self[self.sim]['s'][0:Ns] + self[self.sim]['s'][1 : Ns + 1]) * 0.5)
        JM1new = copy.deepcopy(JM1)
        JM1new[1:-1, :] = griddata(x.flatten(), JM1[0:-1, :], np.asarray(self[self.sim]['s'][1 : Ns - 1]).flatten(), method='cubic')
        JM1new[0, :] = 0
        JM1new[-1, :] = JM1new[-2, :]
        JM1 = copy.deepcopy(JM1new)

        # Patch first three points
        JM1[0:3] = 0.0
        JM2[0:3] = 0.0
        JM3[0:3] = 0.0

        # Remove surface currents
        if True:
            II = np.where((self[self.sim]['s'] > 0.9985) & (self[self.sim]['s'] <= 1.0))
            JM1[II[0][:]] = 0
            JM2[II[0][:]] = 0
            JM3[II[0][:]] = 0

        #  get phase of JM2 for m=2 harmonic at rational surface Iratsurf(1)
        #  and modify phase to remove real part
        if len(self[self.sim]['Iratsurf']):
            from cmath import log

            m0 = 2
            I0 = np.int(self[self.sim]['Iratsurf'][0])
            m1 = np.int(m0 - Mm[0] + 1)
            p0 = -np.angle(JM2[I0 - 1, m1 - 1]) + np.pi / 2.0
            f0 = np.exp(1j * p0)
            JM1 = JM1 * f0
            JM2 = JM2 * f0
            JM3 = JM3 * f0
        self['JPLASMA'] = xarray.Dataset()
        self['JPLASMA']['JM1'] = xarray.DataArray(JM1, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'][:Ns], 'Mm': Mm})
        self['JPLASMA']['JM2'] = xarray.DataArray(JM2, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'][:Ns], 'Mm': Mm})
        self['JPLASMA']['JM3'] = xarray.DataArray(JM3, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'][:Ns], 'Mm': Mm})

    def read_BPLASMA(self, filename, BNORM=1.0):
        """
        Read BPLASMA.OUT file corresponding to perturbed magnetic field

        :param filename: name of file

        :param BNORM: normalization
        """
        spline_B23 = 1
        BPLASMA = np.loadtxt(filename)
        Nm1, Ns, n, _, _, _ = map(int, BPLASMA[0, :])
        BPLASMA = BPLASMA[1:, :]

        Mm = BPLASMA[:Nm1, 1].astype(int)

        BM1 = BPLASMA[Nm1:, 0] + 1j * BPLASMA[Nm1:, 1]
        BM2 = BPLASMA[Nm1:, 2] + 1j * BPLASMA[Nm1:, 3]
        BM3 = BPLASMA[Nm1:, 4] + 1j * BPLASMA[Nm1:, 5]

        BM1 = np.reshape(BM1, (Nm1, Ns)).T
        BM2 = np.reshape(BM2, (Nm1, Ns)).T
        BM3 = np.reshape(BM3, (Nm1, Ns)).T

        BM1 = BM1[0:Ns, :] * BNORM
        BM2 = BM2[0:Ns, :] * BNORM
        BM3 = BM3[0:Ns, :] * BNORM
        if spline_B23 == 2:
            BM2[1:, :] = BM2[0:-1, :]
            BM3[1:, :] = BM3[0:-1, :]
        elif spline_B23 == 1:
            x = np.asarray((self[self.sim]['s'][0 : Ns - 1] + self[self.sim]['s'][1:Ns]) * 0.5)
            # B2
            BM2new = copy.deepcopy(BM2)
            BM2new[1:-1, :] = griddata(x.flatten(), BM2[0:-1, :], np.asarray(self[self.sim]['s'][1 : Ns - 1]).flatten(), method='cubic')
            BM2new[0, :] = 0
            BM2new[-1, :] = BM2new[-2, :]
            BM2 = copy.deepcopy(BM2new)
            # B3
            BM3new = copy.deepcopy(BM3)
            BM3new[1:-1, :] = griddata(x.flatten(), BM3[0:-1, :], np.asarray(self[self.sim]['s'][1 : Ns - 1]).flatten(), method='cubic')
            BM3new[0, :] = 0
            BM3new[-1, :] = BM3new[-2, :]
            BM3 = copy.deepcopy(BM3new)
        self['BPLASMA'] = xarray.Dataset()
        self['BPLASMA']['BM1'] = xarray.DataArray(BM1, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'][:Ns], 'Mm': Mm})
        self['BPLASMA']['BM2'] = xarray.DataArray(BM2, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'][:Ns], 'Mm': Mm})
        self['BPLASMA']['BM3'] = xarray.DataArray(BM3, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'][:Ns], 'Mm': Mm})

    def read_XPLASMA(self, filename, XNORM=1.0):
        """
        Read XPLASMA.OUT file corresponding to perturbed plasma displacement

        :param filename: name of file

        :param XNORM: normalization
        """
        XPLASMA = np.loadtxt(filename)
        # Here Ns is the number of grid points in plasma (what elsewhere is Ns1)
        Nm1, Ns, n, _, _, _ = map(int, XPLASMA[0, :])
        XPLASMA = XPLASMA[1:, :]
        Mm = XPLASMA[:Nm1, 1].astype(int)
        #
        dPSIds = XPLASMA[Nm1 : Nm1 + Ns, 0]
        T = XPLASMA[Nm1 : Nm1 + Ns, 3]
        XM1 = XPLASMA[Nm1 + Ns :, 0] + 1j * XPLASMA[Nm1 + Ns :, 1]
        XM2 = XPLASMA[Nm1 + Ns :, 2] + 1j * XPLASMA[Nm1 + Ns :, 3]
        XM3 = XPLASMA[Nm1 + Ns :, 4] + 1j * XPLASMA[Nm1 + Ns :, 5]
        XM1 = np.reshape(XM1, (Nm1, Ns)).T
        XM2 = np.reshape(XM2, (Nm1, Ns)).T
        XM3 = np.reshape(XM3, (Nm1, Ns)).T
        XM1 = XM1[0:Ns, :] * XNORM
        XM2 = XM2[0:Ns, :] * XNORM
        XM3 = XM3[0:Ns, :] * XNORM
        # Recalculate XM2 and XM3 at integer points (they are defined at half-points)
        x = np.asarray((self[self.sim]['s'][0 : Ns - 1] + self[self.sim]['s'][1:Ns]) * 0.5)
        # X2
        XM2new = copy.deepcopy(XM2)
        XM2new[1:-1, :] = griddata(x.flatten(), XM2[0:-1, :], np.asarray(self[self.sim]['s'][1 : Ns - 1]).flatten(), method='cubic')
        XM2new[0, :] = 0
        XM2new[-1, :] = XM2new[-2, :]
        XM2 = copy.deepcopy(XM2new)
        # X3
        XM3new = copy.deepcopy(XM3)
        XM3new[1:-1, :] = griddata(x.flatten(), XM3[0:-1, :], np.asarray(self[self.sim]['s'][1 : Ns - 1]).flatten(), method='cubic')
        XM3new[0, :] = 0
        XM3new[-1, :] = XM3new[-2, :]
        XM3 = copy.deepcopy(XM3new)
        # change central points of XM2 (see MacReadVPLASMA)
        XM2[:, 0] = XM2[:, 2]
        XM2[:, 1] = XM2[:, 2]
        self['XPLASMA'] = xarray.Dataset()
        # Displacement only defined within plasma, radial coordinate is 'sp'
        self['XPLASMA']['XM1'] = xarray.DataArray(XM1, dims=['sp', 'Mm'], coords={'sp': self[self.sim]['sp'], 'Mm': Mm})
        self['XPLASMA']['XM2'] = xarray.DataArray(XM2, dims=['sp', 'Mm'], coords={'sp': self[self.sim]['sp'], 'Mm': Mm})
        self['XPLASMA']['XM3'] = xarray.DataArray(XM3, dims=['sp', 'Mm'], coords={'sp': self[self.sim]['sp'], 'Mm': Mm})
        self[self.sim]['dPSIds'] = xarray.DataArray(dPSIds, dims=['sp'], coords={'sp': self[self.sim]['sp']})
        self[self.sim]['T'] = xarray.DataArray(T, dims=['sp'], coords={'sp': self[self.sim]['sp']})

    def read_RESULTS(self, filename):
        with open(filename) as fin:
            tmp = list(map(float, fin.readline().split()))
        if len(tmp):
            self[self.sim]['NITR'] = xarray.DataArray(tmp[1])
            self[self.sim]['TALPHA1'] = xarray.DataArray(complex(tmp[2], tmp[3]))
            self[self.sim]['EIG'] = xarray.DataArray(complex(tmp[4], tmp[5]), attrs={'Description': 'Output eigenvalue', 'units': '-'})

    def read_PROFEQ(self, filename):
        """
        Read equilibrium quantities used for MARS calculation
        Returns profiles in physical units, not MARS normalization

        :param filename: name of MARS output file, usually PROFEQ.OUT

        """
        mu0 = constants.mu_0
        R0 = self[self.sim]['R0EXP'].values.astype(float)
        B0 = self[self.sim]['B0EXP'].values.astype(float)

        data = np.loadtxt(filename)

        SY = [
            ('s_eq', 's equilibrium (no wall included)', None),
            ('q_prof', 'safety factor', None),
            ('j_prof', 'current density', B0 / (mu0 * R0)),
            ('p_prof', 'pressure', B0 * B0 / mu0),
            ('n_prof', 'density', None),
            ('v_prof', 'toroidal rotation', None),
            ('eta', 'resistivity=tau_A / tau_R = 1/S', None),
            ('Gamma', 'Gamma', None),
            ('visc_prof', 'viscosity', None),
            ('Ti_prof', 'Ti', None),
            ('Te_prof', 'Te', None),
            ('dpsi/ds', 'dpsi/ds', None),
            ('F', 'F', None),
            ('omega_*i', 'ion dia. drift freq.', None),
            ('omega_*e', 'e- dia. drift freq.', None),
            ('CSV', 'CSV', None),
            ('erot_prof', 'electron rotation', None),
        ]

        x = data[:, 0]
        self['PROFEQ'] = xarray.Dataset()
        for k, (name, des, units) in enumerate(SY):
            if k < data.shape[1]:
                y = np.array(data[:, k])
                if units is not None:
                    y = y * units
                self['PROFEQ'][name] = xarray.DataArray(y, dims=['s_eq'], coords={'s_eq': x}, attrs={'Description': des})
            else:
                printw('%s information is missing from %s file' % (des, os.path.split(filename)[1]))

    def read_FREQS(self, filename):
        """
        Reads all frequencies related to drift kinetic resonances

        :param filename: name of MARS output file, usually FREQUENCIES.OUT
        """
        data = np.loadtxt(filename)
        s = data[:, 0]
        we = data[:, 1]
        wsni = data[:, 2]
        wsne = data[:, 3]
        wsti = data[:, 4]
        wste = data[:, 5]
        awbp = data[:, 6]
        awbt = data[:, 7]
        awdi = data[:, 8]
        awdi[0] = awdi[2]
        awdi[1] = awdi[2]
        awde = data[:, 9]
        awda = data[:, 10]
        awda[0] = awda[2]
        awda[1] = awda[2]
        fracp = data[:, 11]
        fract = data[:, 12]

        lam1 = data[:, 13]
        lam0 = data[:, 14]
        lam2 = data[:, 15]
        awda1 = data[:, 16]
        awda2 = data[:, 17]

        # Storing data
        self['FREQUENCIES'] = xarray.Dataset()
        self['FREQUENCIES']['s'] = xarray.DataArray(s, dims=['s'], coords={'s': s}, attrs={'Description': 's-coord for FREQUENCIES'})
        self['FREQUENCIES']['we'] = xarray.DataArray(we, dims=['s'], coords={'s': s}, attrs={'Description': 'ExB velocity'})
        self['FREQUENCIES']['wsni'] = xarray.DataArray(
            wsni, dims=['s'], coords={'s': s}, attrs={'Description': 'Diamagnetic frequency - density gradient'}
        )
        self['FREQUENCIES']['wsti'] = xarray.DataArray(
            wsti, dims=['s'], coords={'s': s}, attrs={'Description': 'Diamagnetic frequency - temp. gradient'}
        )
        self['FREQUENCIES']['awbt'] = xarray.DataArray(awbt, dims=['s'], coords={'s': s}, attrs={'Description': 'Bounce - trapped'})
        self['FREQUENCIES']['awbp'] = xarray.DataArray(awbp, dims=['s'], coords={'s': s}, attrs={'Description': 'Bounce - passing'})
        self['FREQUENCIES']['awdi'] = xarray.DataArray(awdi, dims=['s'], coords={'s': s}, attrs={'Description': 'Precession - th ions'})
        self['FREQUENCIES']['awde'] = xarray.DataArray(awde, dims=['s'], coords={'s': s}, attrs={'Description': 'Precession - electrons'})
        self['FREQUENCIES']['awda'] = xarray.DataArray(awda, dims=['s'], coords={'s': s}, attrs={'Description': 'Precession - hot'})

    def read_TIMEEVOL(self, filename, ncase, NORM=False):
        """
        Reads time evolution output from time-stepping runs

        :param filename: name of MARS output file, usually TIMEEVOL.OUT

        :param ncase: input namelist variable defining type of MARS-* run

        :param NORM: toggle normalization to physical units
        """
        if NORM:
            mu0 = constants.mu_0
            R0 = self[self.sim]['R0EXP'].values.astype(float)
            B0 = self[self.sim]['B0EXP'].values.astype(float)
            TAUA0 = 1.0e-7  # Need to parse log_mars to get TAUA0!
            xnorm = R0 * 1.0e3  # displacement [mm]
            bnorm = B0 * 1.0e4  # b-field [Gauss]
            cnorm = R0 * B0 / mu0  # current [A]
            tdnorm = (B0 ** 2) / mu0 / np.pi  # torque density [N/m^2]
            ttnorm = tdnorm * (R0 ** 3)  # total torque [Nm]
            tmnorm = TAUA0 * 1.0e3  # time [ms]
            unit_t = 'ms'
            unit_b = 'Gauss'
            unit_c = 'A'
            unit_v = 'A.U.'  # Implement voltage normalization
        else:
            xnorm = 1.0
            bnorm = 1.0
            cnorm = 1.0
            tdnorm = 1.0
            ttnorm = 1.0
            tmnorm = 1.0
            unit_t = 'A.U.'
            unit_b = 'A.U.'
            unit_c = 'A.U.'
            unit_v = 'A.U.'  # Implement voltage normalization
        # tmp = np.atleast_2d(loadtxt(filename, skiprows=0))
        tmp = np.loadtxt(filename)
        ncols = tmp.shape[1]
        self['TIMEEVOL'] = xarray.Dataset()
        if ncase == 3:
            t = np.cumsum(tmp[:, 0]) * tmnorm
            self['TIMEEVOL']['t'] = xarray.DataArray(
                t, dims=['t'], coords={'t': t}, attrs={'Description': 'Time axis', 'Units': '%s' % unit_t}
            )
            # Point-like sensor signal
            self['TIMEEVOL']['PSIS_re'] = xarray.DataArray(
                tmp[:, 1] * bnorm, dims=['t'], coords={'t': t}, attrs={'Description': 'Sensor signal (Re)', 'Units': '%s' % unit_b}
            )
            self['TIMEEVOL']['PSIS_im'] = xarray.DataArray(
                tmp[:, 2] * bnorm, dims=['t'], coords={'t': t}, attrs={'Description': 'Sensor signal (Im)', 'Units': '%s' % unit_b}
            )
            # Get number of columns where AIF is stored
            midcols = tmp.shape[1] - 3
            ncoil = midcols // 4
            # Feedback current
            istart = 3
            for ii in range(ncoil):
                self['TIMEEVOL']['AIF_%d_re' % ii] = xarray.DataArray(
                    tmp[:, istart + ii],
                    dims=['t'],
                    coords={'t': t},
                    attrs={'Description': 'Feedback current coil %s (Re)' % str(ii + 1), 'Units': '%s' % unit_c},
                )
                self['TIMEEVOL']['AIF_%d_im' % ii] = xarray.DataArray(
                    tmp[:, istart + ncoil + ii],
                    dims=['t'],
                    coords={'t': t},
                    attrs={'Description': 'Feedback current coil %s (Im)' % str(ii + 1), 'Units': '%s' % unit_c},
                )
        elif ncase == 4:
            t = np.cumsum(tmp[:, 0]) * tmnorm
            self['TIMEEVOL']['t'] = xarray.DataArray(
                t, dims=['t'], coords={'t': t}, attrs={'Description': 'Time axis', 'Units': '%s' % unit_t}
            )
            # Point-like sensor signal
            self['TIMEEVOL']['PSIS_re'] = xarray.DataArray(
                tmp[:, 1] * bnorm, dims=['t'], coords={'t': t}, attrs={'Description': 'Sensor signal (Re)', 'Units': '%s' % unit_b}
            )
            self['TIMEEVOL']['PSIS_im'] = xarray.DataArray(
                tmp[:, 2] * bnorm, dims=['t'], coords={'t': t}, attrs={'Description': 'Sensor signal (Im)', 'Units': '%s' % unit_b}
            )
            # Sensor noise
            self['TIMEEVOL']['PSISNOISE_re'] = xarray.DataArray(
                tmp[:, -2] * bnorm, dims=['t'], coords={'t': t}, attrs={'Description': 'Sensor noise (Re)', 'Units': '%s' % unit_b}
            )
            self['TIMEEVOL']['PSISNOISE_im'] = xarray.DataArray(
                tmp[:, -1] * bnorm, dims=['t'], coords={'t': t}, attrs={'Description': 'Sensor noise (Im)', 'Units': '%s' % unit_b}
            )
            # Get number of columns where AVF & AIF are stored
            midcols = tmp.shape[1] - 5
            ncoil = midcols // 4
            # Feedback voltages
            istart = 3
            for ii in range(ncoil):
                self['TIMEEVOL']['AVF_%d_re' % ii] = xarray.DataArray(
                    tmp[:, istart + ii],
                    dims=['t'],
                    coords={'t': t},
                    attrs={'Description': 'Feedback voltage coil %s (Re)' % str(ii + 1), 'Units': '%s' % unit_v},
                )
                self['TIMEEVOL']['AVF_%d_im' % ii] = xarray.DataArray(
                    tmp[:, istart + ncoil + ii],
                    dims=['t'],
                    coords={'t': t},
                    attrs={'Description': 'Feedback voltage coil %s (Im)' % str(ii + 1), 'Units': '%s' % unit_v},
                )
            # Feedback currents
            istart = 3 + 2 * ncoil
            for ii in range(ncoil):
                self['TIMEEVOL']['AIF_%d_re' % ii] = xarray.DataArray(
                    tmp[:, istart + ii],
                    dims=['t'],
                    coords={'t': t},
                    attrs={'Description': 'Feedback current coil %s (Re)' % str(ii + 1), 'Units': '%s' % unit_c},
                )
                self['TIMEEVOL']['AIF_%d_im' % ii] = xarray.DataArray(
                    tmp[:, istart + ncoil + ii],
                    dims=['t'],
                    coords={'t': t},
                    attrs={'Description': 'Feedback current coil %s (Im)' % str(ii + 1), 'Units': '%s' % unit_c},
                )
        else:
            print('NCASE=%s not implemented yet' % ncase)
            return

    def read_dWk_den(self, filename):
        """
        Reads perturbed kinetic energy density

        :param filename: name of MARS output file, usually DWK_ENERGY_DENSITY.OUT
        """
        data = np.loadtxt(filename, comments='%')
        NR = len(data[:, 0])
        n = len(data[0, :]) // 10
        blocks = np.split(data, n, axis=1)
        self['DWK_ENERGY_DENSITY'] = xarray.Dataset()
        self['DWK_ENERGY_DENSITY']['FULL_MATRIX'] = xarray.DataArray(data)
        for l in range(len(blocks)):
            CSM = blocks[l][:, 2]
            KP = int(blocks[l][0, 0])
            I = int(blocks[l][1, 1])
            self['DWK_ENERGY_DENSITY']['%s_%s_REdwppara' % (KP, I)] = xarray.DataArray(
                blocks[l][:, 4],
                dims=['s'],
                coords={'s': CSM},
                attrs={'Description': 'Real parallel dWk for species = %s, type = %s' % (KP, I)},
            )
            self['DWK_ENERGY_DENSITY']['%s_%s_IMdwppara' % (KP, I)] = xarray.DataArray(
                blocks[l][:, 5],
                dims=['s'],
                coords={'s': CSM},
                attrs={'Description': f'Imag parallel dWk for species = {KP}, type = {I}'},
            )
            self['DWK_ENERGY_DENSITY']['%s_%s_REdwpperp' % (KP, I)] = xarray.DataArray(
                blocks[l][:, 6],
                dims=['s'],
                coords={'s': CSM},
                attrs={'Description': 'Real perp. dWk for specie = %s, type = %s' % (KP, I)},
            )
            self['DWK_ENERGY_DENSITY']['%s_%s_IMdwpperp' % (KP, I)] = xarray.DataArray(
                blocks[l][:, 7],
                dims=['s'],
                coords={'s': CSM},
                attrs={'Description': 'Imag. perp. dWk for specie = %s, type = %s' % (KP, I)},
            )
            self['DWK_ENERGY_DENSITY']['%s_%s_REdwk' % (KP, I)] = xarray.DataArray(
                blocks[l][:, 8],
                dims=['s'],
                coords={'s': CSM},
                attrs={'Description': 'Real total dWk for specie = %s, type = %s' % (KP, I)},
            )
            self['DWK_ENERGY_DENSITY']['%s_%s_IMdwk' % (KP, I)] = xarray.DataArray(
                blocks[l][:, 9],
                dims=['s'],
                coords={'s': CSM},
                attrs={'Description': 'Imag total dWk for specie = %s, type = %s' % (KP, I)},
            )
        return

    """
    'get' methods, calculate quantities from previously read data
    """

    def get_RZ(self):
        """
        convert RM and ZM into R and Z real space co-ordinates
        """
        Nm2 = self[self.sim]['Nm0'].values.astype(int)
        Nchi = self[self.sim]['Nchi'].values.astype(int)
        chi = np.linspace(np.pi * -1, np.pi, Nchi)
        m = np.arange(0, Nm2, 1)
        m = np.reshape(m, (len(m), 1))
        chi = np.reshape(chi, (len(chi), 1))
        expmchi = np.exp(m * chi.T * 1j)
        R = np.real(np.dot(self[self.sim]['RM'][:, 0:Nm2], expmchi))
        Z = np.real(np.dot(self[self.sim]['ZM'][:, 0:Nm2], expmchi))
        self[self.sim]['R'] = xarray.DataArray(
            R, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        self[self.sim]['Z'] = xarray.DataArray(
            Z, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        return R, Z

    def get_SurfS(self, rs, saveCarMadata=False):
        """
        Generate MacSurfS ASCII file containing control surface for Equivalent Surface Current workflow
        :param rs: radius (r/a) of control surface picked from CHEASE vacuum mesh
        :param saveCarMadata: flag to save MacDataS for CarMa coupling
        """
        if 'R' not in self[self.sim]:
            self.get_RZ()
        II = np.argmin(abs(self[self.sim]['s'].values - rs))
        tmp1 = np.array([self[self.sim]['R'].values[II - 1, :], self[self.sim]['Z'].values[II - 1, :]]).T
        tmp2 = np.array([self[self.sim]['R'].values[II, :], self[self.sim]['Z'].values[II, :]]).T
        RZrw_CarMa = np.array([(tmp1 + tmp2) / (2 * self[self.sim]['R0EXP'].values), tmp2 * self[self.sim]['R0EXP'].values])
        RZrw_EF = tmp2 * self[self.sim]['R0EXP'].values
        # Save and load into tree
        np.savetxt('MacSurfS', RZrw_EF)
        self['CouplingSurfS'] = {}
        self['CouplingSurfS']['MacSurfS'] = OMFITascii('./MacSurfS')
        self['CouplingSurfS']['Rs'] = self[self.sim]['R'].values[II, :]
        self['CouplingSurfS']['Zs'] = self[self.sim]['Z'].values[II, :]
        self['CouplingSurfS']['s'] = rs
        if saveCarMadata:
            np.savetxt('MacDataS', RZrw_CarMa)
            self['CouplingDataS'] = {}
            self['CouplingDataS']['MacDataS'] = OMFITascii('./MacDataS')
            self['CouplingDataS']['RZrw_CarMa'] = RZrw_CarMa
            return RZrw_EF, RZrw_CarMa
        else:
            return RZrw_EF

    def get_UnitVec(self, vacFlag=False):
        """
        Get unit vectors e_s and e_chi and jacobian from real space R,Z
        :param vacFlag: flag to calculate metric elements in all domain (plasma+vacuum)
        """
        from scipy.interpolate import InterpolatedUnivariateSpline

        if 'R' not in self[self.sim]:
            self.get_RZ()
        Ls = len(self[self.sim]['R'][:, 0])
        Lchi = len(self[self.sim]['R'][0, :])
        dRds = copy.deepcopy(self[self.sim]['R'].values)
        dZds = copy.deepcopy(self[self.sim]['Z'].values)
        dRdchi = copy.deepcopy(self[self.sim]['R'].values)
        dZdchi = copy.deepcopy(self[self.sim]['Z'].values)
        jacobian = copy.deepcopy(self[self.sim]['R'].values)
        # Number of mesh points in plasma
        Ns1 = self[self.sim]['Ns1'].values.astype(int)
        # Number of mesh points in vacuum
        Ns2 = self[self.sim]['Ns2'].values.astype(int)
        # indexes of full domain: first couple is vacuum, second is plasma
        domain = [(Ns1, Ns1 + Ns2), (0, Ns1)]
        # Loop over vacuum and plasma
        for ii, kk in domain:
            if ii == Ns1:
                printi('Partial derivatives in vacuum')
                s0 = self[self.sim]['svac'].values
            elif ii == 0:
                printi('Partial derivatives in plasma')
                s0 = self[self.sim]['sp'].values
            else:
                printi('ERROR! Check Jacobian calculation')
                return
            II1 = ii
            II2 = kk
            R0 = self[self.sim]['R'].isel(s=range(II1, II2))
            Z0 = self[self.sim]['Z'].isel(s=range(II1, II2))
            Nchi = self[self.sim]['Nchi'].values.astype(int)
            chi0 = np.linspace(np.pi * -1, np.pi, Nchi)
            hs = 0.5 * ((s0[1:] - s0[:-1]).min())
            hs = min(hs, 2e-5)
            hchi = 0.5 * ((chi0[1:] - chi0[:-1]).min())
            hchi = min(hchi, 1e-4)
            s1 = s0 - hs
            s2 = s0 + hs
            chi1 = chi0 - hchi
            chi2 = chi0 + hchi

            # compute dR/ds using R(s,chi) and spline
            R1 = np.zeros(np.shape(R0))
            R2 = np.zeros(np.shape(R0))
            for k in range(len(R0[0, :])):
                R1[:, k] = InterpolatedUnivariateSpline(s0, R0[:, k], bbox=[s1[0], s0[-1]])(s1)
                R2[:, k] = InterpolatedUnivariateSpline(s0, R0[:, k], bbox=[s0[0], s2[-1]])(s2)
            dRds[II1:II2, :] = (R2 - R1) / (2.0 * hs)

            # compute dZ/ds using Z(s,chi) and spline
            Z1 = np.zeros(np.shape(Z0))
            Z2 = np.zeros(np.shape(Z0))
            for k in range(len(R0[0, :])):
                Z1[:, k] = InterpolatedUnivariateSpline(s0, Z0[:, k], bbox=[s1[0], s0[-1]])(s1)
                Z2[:, k] = InterpolatedUnivariateSpline(s0, Z0[:, k], bbox=[s0[0], s2[-1]])(s2)
            dZds[II1:II2, :] = (Z2 - Z1) / (2.0 * hs)

            # compute dR/dchi using R(s,chi) and spline
            R1 = np.zeros(np.shape(R0))
            R2 = np.zeros(np.shape(R0))
            for k in range(len(R0[:, 0])):
                R1[k, :] = InterpolatedUnivariateSpline(chi0, R0[k, :], bbox=[chi1[0], chi0[-1]])(chi1)
                R2[k, :] = InterpolatedUnivariateSpline(chi0, R0[k, :], bbox=[chi0[0], chi2[-1]])(chi2)
            R1[:, 0] = R1[:, -1]
            R2[:, -1] = R2[:, 0]
            dRdchi[II1:II2, :] = (R2 - R1) / (2.0 * hchi)

            # compute dR/dchi using R(s,chi) and spline
            Z1 = np.zeros(np.shape(Z0))
            Z2 = np.zeros(np.shape(Z0))
            for k in range(len(R0[:, 0])):
                Z1[k, :] = InterpolatedUnivariateSpline(chi0, Z0[k, :], bbox=[chi1[0], chi0[-1]])(chi1)
                Z2[k, :] = InterpolatedUnivariateSpline(chi0, Z0[k, :], bbox=[chi0[0], chi2[-1]])(chi2)
            Z1[:, 0] = Z1[:, -1]
            Z2[:, -1] = Z2[:, 0]
            dZdchi[II1:II2, :] = (Z2 - Z1) / (2.0 * hchi)
        # Calculate Jacobian
        J = (self[self.sim]['R'].values) * (-dRdchi * dZds + dRds * dZdchi)
        J[0, :] = J[1, :]
        if vacFlag:
            # Calculate metric elements in both plasma and vacuum
            G11 = np.square(dRds) + np.square(dZds)
            G12 = np.multiply(dRds, dRdchi) + np.multiply(dZds, dZdchi)
            G22 = np.square(dRdchi) + np.square(dZdchi)
            G22[0, :] = G22[1, :]
            G33 = np.square(self[self.sim]['R'].values)
        else:
            # Calculate metric elements inside the plasma
            G11 = np.square(dRds[0:Ns1, :]) + np.square(dZds[0:Ns1, :])
            G12 = np.multiply(dRds[0:Ns1, :], dRdchi[0:Ns1, :]) + np.multiply(dZds[0:Ns1, :], dZdchi[0:Ns1, :])
            G22 = np.square(dRdchi[0:Ns1, :]) + np.square(dZdchi[0:Ns1, :])
            G22[0, :] = G22[1, :]
            G33 = np.square(self[self.sim]['R'].values[0:Ns1, :])
        # Store data
        self['UnitVector'] = xarray.Dataset()
        self['UnitVector']['dRds'] = xarray.DataArray(
            dRds, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        self['UnitVector']['dZds'] = xarray.DataArray(
            dZds, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        self['UnitVector']['dRdchi'] = xarray.DataArray(
            dRdchi, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        self['UnitVector']['dZdchi'] = xarray.DataArray(
            dZdchi, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        self['UnitVector']['Jacobian'] = xarray.DataArray(
            J, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
        )
        self['Metric'] = xarray.Dataset()
        if vacFlag:
            self['Metric']['G11'] = xarray.DataArray(
                G11, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
            )
            self['Metric']['G12'] = xarray.DataArray(
                G12, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
            )
            self['Metric']['G22'] = xarray.DataArray(
                G22, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
            )
            self['Metric']['G33'] = xarray.DataArray(
                G33, dims=['s', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 's': self[self.sim]['s']}
            )
        else:
            # Radial coordinate name 'sp' is within plasma
            self['Metric']['G11'] = xarray.DataArray(
                G11, dims=['sp', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 'sp': self[self.sim]['sp']}
            )
            self['Metric']['G12'] = xarray.DataArray(
                G12, dims=['sp', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 'sp': self[self.sim]['sp']}
            )
            self['Metric']['G22'] = xarray.DataArray(
                G22, dims=['sp', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 'sp': self[self.sim]['sp']}
            )
            self['Metric']['G33'] = xarray.DataArray(
                G33, dims=['sp', 'chi'], coords={'chi': np.linspace(np.pi * -1, np.pi, Nchi), 'sp': self[self.sim]['sp']}
            )
        return dRds, dZds, dRdchi, dZdchi, jacobian

    def get_X123(self):
        """
        Inverse Fourier transorm of plasma displacement and projection along Chi and Phi
        """
        if 'UnitVector' not in self:
            self.get_UnitVec()
        expmchi = np.exp(self['XPLASMA']['Mm'] * self[self.sim]['chi'] * 1j)
        expmchi_inv = np.exp(-self[self.sim]['chi'] * self['XPLASMA']['Mm'] * 1j)
        chione = np.ones((1, len(self[self.sim]['chi'])))
        Ns1 = self[self.sim]['Ns1'].values
        X1a = np.dot(self['XPLASMA']['XM1'].values, expmchi)
        X2a = np.dot(self['XPLASMA']['XM2'].values, expmchi)
        X3a = np.dot(self['XPLASMA']['XM3'].values, expmchi)
        if np.size(self['Metric']['G22'], 0) != np.size(X1a, 0):
            self.get_UnitVec(vacFlag=False)
        # Calculate X1 along e_s
        X1 = X1a
        # Calculate X2 along e_chi
        dPSIds = np.array(self[self.sim]['dPSIds'].values).reshape((len(self[self.sim]['dPSIds'].values), 1))
        J = self['UnitVector']['Jacobian'].isel(s=range(Ns1)).values
        T = np.array(self[self.sim]['T'].values).reshape((len(self[self.sim]['T']), 1))
        Bchi = np.divide(np.dot(dPSIds, chione), J)
        Bphi = np.divide(np.dot(T, chione), np.square(self[self.sim]['R'].values[0:Ns1, :]))
        B2 = np.multiply(np.square(Bchi), self['Metric']['G22']) + np.multiply(np.square(Bphi), self['Metric']['G33'])
        X2 = (
            np.divide(np.multiply(-X1, np.multiply(self['Metric']['G12'], np.square(Bchi))), B2)
            + np.divide(np.multiply(X2a, np.multiply(Bphi, self['Metric']['G33'])), B2)
            - np.multiply(X3a, Bchi)
        )
        # Calculate X3 along e_phi
        X3 = (
            np.divide(np.multiply(-X1, np.multiply(self['Metric']['G12'], np.multiply(Bchi, Bphi))), B2)
            + np.divide(np.multiply(X2a, np.multiply(Bchi, self['Metric']['G22'])), B2)
            - np.multiply(X3a, Bphi)
        )
        # Calculate normal plasma displacement
        Xn = np.divide(np.multiply(X1a, J), np.sqrt(np.multiply(self['Metric']['G33'], self['Metric']['G22'])))
        # Fourier transorm of Xn
        XnM = np.dot(Xn, expmchi_inv) * (self[self.sim]['chi'].values[1] - self[self.sim]['chi'].values[0]) / (2 * np.pi)
        # Radial coordinate name 's' is retained for compatibility, here 's' is 'sp' (within plasma)
        self['XPLASMA']['Bchi'] = xarray.DataArray(
            Bchi, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']}
        )
        self['XPLASMA']['Bphi'] = xarray.DataArray(
            Bphi, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']}
        )
        self['XPLASMA']['B2'] = xarray.DataArray(B2, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']})
        self['XPLASMA']['X1'] = xarray.DataArray(X1, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']})
        self['XPLASMA']['X2'] = xarray.DataArray(X2, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']})
        self['XPLASMA']['X3'] = xarray.DataArray(X3, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']})
        self['XPLASMA']['Xn'] = xarray.DataArray(Xn, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']})
        self['XPLASMA']['XnM'] = xarray.DataArray(XnM, dims=['sp', 'Mm'], coords={'sp': self[self.sim]['sp'], 'Mm': self['XPLASMA']['Mm']})
        self['XPLASMA']['X1a'] = xarray.DataArray(
            X1a, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']}
        )
        self['XPLASMA']['X2a'] = xarray.DataArray(
            X2a, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']}
        )
        self['XPLASMA']['X3a'] = xarray.DataArray(
            X3a, dims=['sp', 'chi'], coords={'sp': self[self.sim]['sp'], 'chi': self[self.sim]['chi']}
        )

    def get_B123(self):
        """
        Inverse Fourier transorm of perturbed magnetic field and calculate Bn
        """
        if 'UnitVector' not in self:
            self.get_UnitVec(vacFlag=True)
        expmchi = np.exp(self['BPLASMA']['Mm'] * self[self.sim]['chi'] * 1j)
        expmchi_inv = np.exp(-self[self.sim]['chi'] * self['BPLASMA']['Mm'] * 1j)
        chione = np.ones((1, len(self[self.sim]['chi'])))
        Ns1 = self[self.sim]['Ns1'].values
        # Calculate inverse Fourier transorm
        B1 = np.dot(self['BPLASMA']['BM1'].values, expmchi)
        B2 = np.dot(self['BPLASMA']['BM2'].values, expmchi)
        B3 = np.dot(self['BPLASMA']['BM3'].values, expmchi)
        # Check metric element definition
        if np.size(self['Metric']['G22'], 0) != np.size(B1, 0):
            self.get_UnitVec(vacFlag=True)
        # Calculate normal perturbeb magnetic field
        Bn = np.divide(B1, np.divide(np.sqrt(self['Metric']['G22']), self[self.sim]['R']))
        # Fourier transorm of Bn
        BnM = np.dot(Bn, expmchi_inv) * (self[self.sim]['chi'].values[1] - self[self.sim]['chi'].values[0]) / (2 * np.pi)
        # Store data
        self['BPLASMA']['B1'] = xarray.DataArray(B1, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})
        self['BPLASMA']['B2'] = xarray.DataArray(B2, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})
        self['BPLASMA']['B3'] = xarray.DataArray(B3, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})
        self['BPLASMA']['Bn'] = xarray.DataArray(Bn, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})
        self['BPLASMA']['BnM'] = xarray.DataArray(BnM, dims=['s', 'Mm'], coords={'s': self[self.sim]['s'], 'Mm': self['BPLASMA']['Mm']})

    def get_Bcyl(self):
        """
        Get B-field components in cylindrical coordinates
        """
        if 'B1' not in self['BPLASMA']:
            self.get_B123()
        if 'UnitVector' not in self:
            self.get_UnitVec(vacFlag=True)
        if 'R' not in self[self.sim]:
            self.get_RZ()
        J = self['UnitVector']['Jacobian']
        dRds = self['UnitVector']['dRds']
        dRdchi = self['UnitVector']['dRdchi']
        dZds = self['UnitVector']['dZds']
        dZdchi = self['UnitVector']['dZdchi']
        R0 = self[self.sim]['R']
        # Calculate Br,z,phi
        Br = (self['BPLASMA']['B1'] * dRds + self['BPLASMA']['B2'] * dRdchi) / J
        Br[0:2, :] = Br[2, :]
        Bz = (self['BPLASMA']['B1'] * dZds + self['BPLASMA']['B2'] * dZdchi) / J
        Bz[0:2, :] = Bz[2, :]
        Bphi = (self['BPLASMA']['B3'] * R0) / J
        Bphi[0:2, :] = Bphi[2, :]
        self['BPLASMA']['Br'] = xarray.DataArray(Br, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})
        self['BPLASMA']['Bz'] = xarray.DataArray(Bz, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})
        self['BPLASMA']['Bphi'] = xarray.DataArray(Bphi, dims=['s', 'chi'], coords={'s': self[self.sim]['s'], 'chi': self[self.sim]['chi']})

    def get_dWk(self, NFIT=3):
        """
        Calculate dWk components from DWK_ENERGY_DENSITY

        :param NFIT: integral of energy density skips the first NFIT+1 points
        """
        if 'DWK_ENERGY_DENSITY' not in self:
            self.read_dWk_den()
        den = self['DWK_ENERGY_DENSITY']['FULL_MATRIX']
        NR = len(den[:, 0])
        n = len(den[0, :]) // 10
        blocks = np.split(den, n, axis=1)
        DWK = np.zeros((n, 8))
        for l in range(len(blocks)):
            REdwppara = 0
            IMdwppara = 0
            REdwpperp = 0
            IMdwpperp = 0
            REdwk = 0
            IMdwk = 0
            # Storing KP (=1...NSPECIES, 0=total)
            DWK[l, 0] = blocks[l][0, 0]
            # Storing I (=1...5)
            DWK[l, 1] = blocks[l][1, 1]
            for j in range(NFIT + 1, NR):
                REdwppara = REdwppara + blocks[l][j, 4] * blocks[l][j, 3]
                IMdwppara = IMdwppara + blocks[l][j, 5] * blocks[l][j, 3]
                REdwpperp = REdwpperp + blocks[l][j, 6] * blocks[l][j, 3]
                IMdwpperp = IMdwpperp + blocks[l][j, 7] * blocks[l][j, 3]
                REdwk = REdwk + blocks[l][j, 8] * blocks[l][j, 3]
                IMdwk = IMdwk + blocks[l][j, 9] * blocks[l][j, 3]
            DWK[l, 2] = REdwppara
            DWK[l, 3] = IMdwppara
            DWK[l, 4] = REdwpperp
            DWK[l, 5] = IMdwpperp
            DWK[l, 6] = REdwk
            DWK[l, 7] = IMdwk
        # Total dWk
        K_tot = []
        # Non-adiabatic Trapped precession Drift
        K_I_NTD = []
        K_e_NTD = []
        # Non-adiabatic Trapped Bounce
        K_I_NTB = []
        K_e_NTB = []
        # Non-adiabatic Passing
        K_I_NP = []
        K_e_NP = []
        species = np.unique(DWK[:, 0])
        Nspecies = len(species) - 1
        self['DWK_COMPONENTS'] = xarray.Dataset()
        self['DWK_COMPONENTS']['NSPECIES'] = Nspecies
        for P in species:
            if P == 0:
                # Storing total kinetic contribution
                (i,) = np.where(DWK[:, 0] == 0.0)
                (j,) = np.where(DWK[i, 1] == 0.0)

                K_tot = complex(DWK[i[j], 6], DWK[i[j], 7])
                self['DWK_COMPONENTS']['K_tot'] = K_tot
            else:
                if P == 1:
                    # ions
                    particle = 'I'
                elif P == 2:
                    # electrons
                    particle = 'e'
                elif P > 2:
                    # Other particle species
                    particle = str(P)
                (l,) = np.where(DWK[:, 0] == P)
                (p,) = np.where(DWK[l, 1] == 3.0)
                (b,) = np.where(DWK[l, 1] == 4.0)
                (d,) = np.where(DWK[l, 1] == 5.0)

                if not p:
                    K_NP = complex(0.0, 0.0)
                else:
                    K_NP = complex(DWK[l[p], 6], DWK[l[p], 7])

                if not b:
                    K_NTB = complex(0.0, 0.0)
                else:
                    K_NTB = complex(DWK[l[b], 6], DWK[l[b], 7])

                if not d:
                    K_NTD = complex(0.0, 0.0)
                else:
                    K_NTD = complex(DWK[l[d], 6], DWK[l[d], 7])
                self['DWK_COMPONENTS']['K_%s_NTD' % particle] = K_NTD
                self['DWK_COMPONENTS']['K_%s_NTB' % particle] = K_NTB
                self['DWK_COMPONENTS']['K_%s_NP' % particle] = K_NP

    @dynaLoad
    def load(self):
        self[self.sim] = xarray.Dataset()
        mapper = OrderedDict(
            (
                ('RMZM_F.OUT', self.read_RMZM),
                ('RMZM_F', self.read_RMZM),
                ('JPLASMA.OUT', self.read_JPLASMA),
                ('JPLASMA', self.read_JPLASMA),
                ('RESULT.OUT', self.read_RESULTS),
                ('RESULT', self.read_RESULTS),
                ('BPLASMA.OUT', self.read_BPLASMA),
                ('BPLASMA', self.read_BPLASMA),
                ('PROFEQ.OUT', self.read_PROFEQ),
                ('PROFEQ', self.read_PROFEQ),
                ('XPLASMA.OUT', self.read_XPLASMA),
                ('XPLASMA', self.read_XPLASMA),
                ('FREQUENCIES.OUT', self.read_FREQS),
                ('DWK_ENERGY_DENSITY.OUT', self.read_dWk_den),
            )
        )
        for item in mapper:
            if os.path.exists(self.filename + os.sep + item):
                mapper[item](self.filename + os.sep + item)

    """
    Plotting methods
    """

    def plot_RMZM(self, fig=None):
        """
        Plot RM-ZM grid
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax1 = fig.add_subplot(2, 2, 1)
        ax1.plot(self[self.sim]['s'], np.real(self[self.sim]['RM']))
        ax1.set_ylabel('Re(R_m)')

        ax2 = fig.add_subplot(2, 2, 2)
        ax2.plot(self[self.sim]['s'], np.imag(self[self.sim]['RM']))
        ax2.set_ylabel('Im(R_m)')

        ax3 = fig.add_subplot(2, 2, 3)
        ax3.plot(self[self.sim]['s'], np.real(self[self.sim]['ZM']))
        ax3.set_ylabel('Re(Z_m)')
        ax3.set_xlabel('s')

        ax4 = fig.add_subplot(2, 2, 4)
        ax4.plot(self[self.sim]['s'], np.imag(self[self.sim]['ZM']))
        ax4.set_ylabel('Im(Z_m)')
        ax4.set_xlabel('s')
        fig.tight_layout()

    def plot_RZ(self, fig=None):
        """
        Plot R-Z grid
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        Ns = self[self.sim]['Ns1'].values.astype(int) + self[self.sim]['Ns2'].values.astype(int)
        Rplot = self[self.sim]['R'][0:Ns, :].values
        Zplot = self[self.sim]['Z'][0:Ns, :].values
        ax.plot(Rplot.T * self[self.sim]['R0EXP'].values, Zplot.T * self[self.sim]['R0EXP'].values, 'g-')
        ax.plot(Rplot * self[self.sim]['R0EXP'].values, Zplot * self[self.sim]['R0EXP'].values, 'b-')
        ax.axis('equal')
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')
        fig.tight_layout()

    def plot_SurfS(self, fig=None):
        """
        Plot surface stored in ['CouplingSurfS']
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        Ns = self[self.sim]['Ns1'].values.astype(int) + self[self.sim]['Ns2'].values.astype(int)
        Ns1 = self[self.sim]['Ns1'].values.astype(int)
        Rtot = self[self.sim]['R'][0:Ns, :].values
        Ztot = self[self.sim]['Z'][0:Ns, :].values
        if 'CouplingSurfS' not in self:
            self.get_SurfS()
        Rs = self['CouplingSurfS']['Rs']
        Zs = self['CouplingSurfS']['Zs']
        Rplasma = self[self.sim]['R'][Ns1, :].values
        Zplasma = self[self.sim]['Z'][Ns1, :].values
        # Plot all surfaces
        ax.plot(Rtot.T * self[self.sim]['R0EXP'].values, Ztot.T * self[self.sim]['R0EXP'].values, color='g')
        # Plot SurfS
        ax.plot(Rs * self[self.sim]['R0EXP'].values, Zs * self[self.sim]['R0EXP'].values, linewidth=3, color='r')
        ax.plot(
            Rplasma * self[self.sim]['R0EXP'].values, Zplasma * self[self.sim]['R0EXP'].values, linestyle='dashed', linewidth=2, color='b'
        )
        ax.axis('equal')
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')
        fig.tight_layout()

    def plot_JPLASMA(self, fig=None):
        """
        Plot perturbed current components J1,J2,J3
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = None
        label = 'JPLASMA'
        for k, c in enumerate(['JM1', 'JM2', 'JM3']):
            ax = ax1 = fig.add_subplot(3, 2, 2 * k + 1, sharex=ax)
            ax1.plot(self[label]['s'], np.real(self[label][c]))
            ax1.set_title('Re($J^%d_m$)' % (k + 1), y=0.8)
            ax2 = fig.add_subplot(3, 2, 2 * (k + 1), sharex=ax)
            ax2.plot(self[label]['s'], np.imag(self[label][c]))
            ax2.set_title('Im($J^%d_m$)' % (k + 1), y=0.8)
        ax1.set_xlabel('s')
        ax2.set_xlabel('s')
        fig.tight_layout()

    def plot_FREQUENCIES(self, fig=None):
        """
        Plot frequencies related to drift kinetic resonances
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        label = 'FREQUENCIES'
        ax.plot(self[label]['s'], self[label]['we'], 'k-', label=r'$\omega_E$')
        ax.plot(self[label]['s'], self[label]['wsni'] + self[label]['wsti'], 'g-', label=r'$\omega_*^{thi}$')
        ax.plot(self[label]['s'], self[label]['awbt'], 'b-', label=r'$\omega_b^{thi}$')
        ax.plot(self[label]['s'], self[label]['awdi'], 'r-', label=r'$\omega_d^{thi}$')
        ax.plot(self[label]['s'], self[label]['awda'], 'r--', label=r'$\omega_d^{hot}$')
        ax.set_xlabel('s')
        ax.set_ylabel('Frequencies')
        ax.legend(loc=0)
        fig.tight_layout()

    def plot_BPLASMA(self, fig=None):
        """
        Plot perturbed field components B1,B2,B3
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = None
        label = 'BPLASMA'
        for k, c in enumerate(['BM1', 'BM2', 'BM3']):
            ax = ax1 = fig.add_subplot(3, 2, 2 * k + 1, sharex=ax)
            ax1.plot(self[label]['s'], np.real(self[label][c]))
            ax1.set_title('Re($B^%d_m$)' % (k + 1), y=0.8)
            ax2 = fig.add_subplot(3, 2, 2 * (k + 1), sharex=ax)
            ax2.plot(self[label]['s'], np.imag(self[label][c]))
            ax2.set_title('Im($B^%d_m$)' % (k + 1), y=0.8)
        ax1.set_xlabel('s')
        ax2.set_xlabel('s')
        fig.tight_layout()

    def plot_XPLASMA(self, fig=None):
        """
        Plot plasma displacement components X1,X2,X3
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = None
        label = 'XPLASMA'
        for k, c in enumerate(['XM1', 'XM2', 'XM3']):
            ax1 = fig.add_subplot(3, 2, 2 * k + 1, sharex=ax)
            ax1.plot(self[label]['sp'], np.real(self[label][c]))
            ax1.set_title('Re($X^%d_m$)' % (k + 1), y=0.8)
            ax2 = fig.add_subplot(3, 2, 2 * (k + 1), sharex=ax)
            ax2.plot(self[label]['sp'], np.imag(self[label][c]))
            ax2.set_title('Im($X^%d_m$)' % (k + 1), y=0.8)
        ax1.set_xlabel('sp')
        ax2.set_xlabel('sp')
        fig.tight_layout()

    def plot_Xn(self, fig=None):
        """
        Plot normal plasma displacement
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        label = 'XPLASMA'
        if 'R' not in self[self.sim]:
            self.get_RZ()
        if 'Xn' not in self[label]:
            self.get_X123()
        ax = fig.add_subplot(1, 1, 1)
        Irat = self[self.sim]['Iratsurf']
        # Plasma displacement defined within plasma boundary
        Ns1 = self[self.sim]['Ns1'].values
        R = self[self.sim]['R'].isel(s=range(Ns1)) * self[self.sim]['R0EXP']
        Z = self[self.sim]['Z'].isel(s=range(Ns1)) * self[self.sim]['R0EXP']
        C = np.real(self[label]['Xn'])
        pc = ax.pcolormesh(R, Z, C, cmap='jet', antialiased=True)
        for I in Irat.values:
            ax.plot(R[I, :], Z[I, :], linestyle='dashed', linewidth=2, color='w')
        ax.set_aspect('equal', adjustable='box')
        fig.colorbar(pc, ax=ax)
        ax.set_title('Re($\\xi_n$)')
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')
        fig.tight_layout()

    def plot_Bn(self, II=False, fig=None):
        """
        Plot normal field perturbation
        :param II: plot up to II radial grid index (default is plasma boundary)
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        label = 'BPLASMA'
        if 'R' not in self[self.sim]:
            self.get_RZ()
        if 'Bn' not in self[label]:
            self.get_B123()
        ax = fig.add_subplot(1, 1, 1)
        Irat = self[self.sim]['Iratsurf']
        # Pick last s-grid point for Bn plotting
        if II:
            Ns = II
        else:
            Ns = self[self.sim]['Ns1'].values
        R = self[self.sim]['R'].isel(s=range(Ns)) * self[self.sim]['R0EXP']
        Z = self[self.sim]['Z'].isel(s=range(Ns)) * self[self.sim]['R0EXP']
        C = np.real(self[label]['Bn'].isel(s=range(Ns)))
        pc = ax.pcolormesh(R, Z, C, cmap='jet', antialiased=True)
        for I in Irat.values:
            ax.plot(R[I, :], Z[I, :], linestyle='dashed', linewidth=2, color='w')
        ax.set_aspect('equal', adjustable='box')
        fig.colorbar(pc, ax=ax)
        ax.set_title('Re($B_n$)')
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')
        fig.tight_layout()

    def plot_BrzSurf(self, II=False, fig=None):
        """
        Plot Br Bz (cyl. coords.) along specified surface
        :param II: plot on II radial grid index (default is plasma boundary)
        :param fig: specify target figure
        """
        # Pick last s-grid point for Bn plotting
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        if II:
            Ns = II
        else:
            Ns = self[self.sim]['Ns1'].values
        R = (self[self.sim]['R'].values) * self[self.sim]['R0EXP'].values
        Z = (self[self.sim]['Z'].values) * self[self.sim]['R0EXP'].values
        Rw = R[Ns, :]
        Zw = Z[Ns, :]
        Tw = np.arctan2(Zw - Z[0, 0], Rw - R[0, 0]) * (180 / np.pi)
        KK = np.where(np.diff(Tw) < -180)[0]
        if (len(KK) == 1) & (2 * KK > len(Tw)):
            Tw[KK[0] + 1 :] = Tw[KK[0] + 1 :] + 360
        if (len(KK) == 1) & (2 * KK < len(Tw)):
            Tw[0 : KK[0]] = Tw[0 : KK[0]] - 360
        Tw = np.sort(Tw)
        JJ = np.argsort(Tw)
        label = 'BPLASMA'
        BR = self[label]['Br'].values
        BZ = self[label]['Bz'].values
        ax1 = fig.add_subplot(2, 2, 1)
        ax1.plot(Tw, np.real(BR[Ns, JJ]), linewidth=2)
        ax1.set_ylabel('Re($\delta B_R$) [Gauss]')

        ax2 = fig.add_subplot(2, 2, 2)
        ax2.plot(Tw, np.imag(BR[Ns, JJ]), linewidth=2)
        ax2.set_ylabel('Im($\delta B_R$) [Gauss]')

        ax3 = fig.add_subplot(2, 2, 3)
        ax3.plot(Tw, np.real(BZ[Ns, JJ]), linewidth=2)
        ax3.set_ylabel('Re($\delta B_Z$) [Gauss]')
        ax3.set_xlabel('GEOM angle')

        ax4 = fig.add_subplot(2, 2, 4)
        ax4.plot(Tw, np.imag(BZ[Ns, JJ]), linewidth=2)
        ax4.set_ylabel('Im($\delta B_Z$) [Gauss]')
        ax4.set_xlabel('GEOM angle')
        fig.tight_layout()

    def plot_BrzAbs(self, fig=None):
        """
        Plot |B| along poloidal angle at plasma surface
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        Ns = self[self.sim]['Ns1'].values
        chi = self[self.sim]['chi'].values
        label = 'BPLASMA'
        AmpB = np.sqrt(abs(self[label]['Br'].values) ** 2 + abs(self[label]['Bz'].values) ** 2 + abs(self[label]['Bphi'].values) ** 2)
        ax.plot(chi * 180 / np.pi, AmpB[Ns, :], linewidth=2)
        ax.set_ylabel('$|\delta B|$ [Gauss/kAt]')
        ax.set_xlabel('Poloidal angle $[\circ]$')
        fig.tight_layout()

    def plot_BrzMap(self, fig=None):
        """
        Plot |B| (R,Z) map inside plasma
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        Ns = self[self.sim]['Ns1'].values
        label = 'BPLASMA'
        ax = fig.add_subplot(1, 1, 1)
        Br = self[label]['Br'].isel(s=range(Ns)).values
        Bz = self[label]['Bz'].isel(s=range(Ns)).values
        Bphi = self[label]['Bphi'].isel(s=range(Ns)).values
        AmpB = np.sqrt(abs(Br) ** 2 + abs(Bz) ** 2 + abs(Bphi) ** 2)
        R = self[self.sim]['R'].isel(s=range(Ns)) * self[self.sim]['R0EXP']
        Z = self[self.sim]['Z'].isel(s=range(Ns)) * self[self.sim]['R0EXP']
        pc = ax.pcolormesh(R, Z, AmpB, cmap='jet', antialiased=True)
        ax.set_aspect('equal', adjustable='box')
        fig.colorbar(pc, ax=ax)
        ax.set_title('$|\delta B|$ [Gauss/kAt]')
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')
        fig.tight_layout()

    def plot_dWkDen(self, fig=None):
        """
        Plot total dWk energy density profiles
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        label = 'DWK_ENERGY_DENSITY'
        csm = self[label]['s']
        dWki_re = self[label]['0_0_REdwk']
        dWki_im = self[label]['0_0_IMdwk']
        ax = fig.add_subplot(1, 1, 1)
        pyplot.plot(csm, dWki_re, 'bo-', label=r'Re($\delta W_k$) total')
        pyplot.plot(csm, dWki_im, 'rs--', label=r'Im($\delta W_k$) total')
        ax.set_xlabel('s')
        ax.set_ylabel('dWk energy density')
        ax.legend(loc=0)
        fig.tight_layout()

    def plot_dWkComp(self, fig=None):
        """
        Plot dWk components (integral of energy density)
        :param fig: specify target figure
        """
        from matplotlib import pyplot

        if fig is None:
            fig = pyplot.figure()
        NSPECIES = self['DWK_COMPONENTS']['NSPECIES'].values
        labels = ['NTD', 'NTB', 'NP']
        ions = [self['DWK_COMPONENTS']['K_I_NTD'], self['DWK_COMPONENTS']['K_I_NTB'], self['DWK_COMPONENTS']['K_I_NP']]
        electrons = [self['DWK_COMPONENTS']['K_e_NTD'], self['DWK_COMPONENTS']['K_e_NTB'], self['DWK_COMPONENTS']['K_e_NP']]
        if NSPECIES > 2:
            SP_DICT = {}
            for II in range(NSPECIES - 2):
                P = II + NSPECIES - 2
                SP_DICT[P] = [
                    self['DWK_COMPONENTS']['K_%s_NTD' % P],
                    self['DWK_COMPONENTS']['K_%s_NTB' % P],
                    self['DWK_COMPONENTS']['K_%s_NP' % P],
                ]
        x = np.arange(len(labels))
        width = 0.35
        ax = fig.add_subplot(1, 1, 1)
        ax.bar(x - width / 2, ions, width, label='Ions')
        ax.bar(x + width / 2, electrons, width, label='Electrons')
        if NSPECIES > 2:
            for II in range(NSPECIES - 2):
                P = II + NSPECIES - 2
                ax.bar(x - width / 2, SP_DICT[P], width, label='Specie %s' % P)
        ax.set_ylabel('dWk')
        ax.set_xticks(x)
        ax.set_xticklabels(labels)
        ax.legend(loc=0)
        fig.tight_layout()


class OMFITnColumns(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with n-columns input files

    :param filename: filename passed to OMFITascii class

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            header = f.readline()
        if not len(header):
            return
        tmp = np.atleast_2d(np.loadtxt(self.filename, skiprows=1))
        self['x'] = tmp[:, 0]
        for k in range(1, tmp.shape[1]):
            self['y_%d' % k] = tmp[:, k]

    @dynaSave
    def save(self):
        with open(self.filename, 'w') as f:
            if not len(self):
                f.write('0 0')
                return
            tmp = np.vstack(list(self.values())).T
            shape = list(tmp.shape)
            shape[1] = shape[1] - 1
            f.write(' '.join(map(str, shape)) + '\n')
            for k in range(tmp.shape[0]):
                f.write(' '.join(['%5.9e' % x for x in tmp[k, :]]) + '\n')

    def plot(self):
        from matplotlib import pyplot

        tmp = np.vstack(list(self.values())).T
        for k in range(tmp.shape[1]):
            # condition to avoid plotting first columnt vs. itself
            if k == 0:
                continue
            pyplot.figure(k)
            pyplot.plot(tmp[:, 0], tmp[:, k])
            pyplot.ylabel(self._OMFITkeyName)
            pyplot.tight_layout()


class OMFITmarsProfile(OMFITnColumns):
    pass


############################################
if '__main__' == __name__:
    test_classes_main_header()

    if os.path.exists('/home/pigatto/Work_2020/RFX-mod2/OUTPUTS/'):
        tmp = OMFITmars('/home/pigatto/Work_2020/RFX-mod2/OUTPUTS/')
        tmp.load()
        print(tmp)
        tmp.plot_RMZM()
        tmp.plot_JPLASMA()
        from matplotlib import pyplot

        pyplot.show()



"""
Functions for adding KSTAR data to the IMAS schema by writing to ODS instances
"""

try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import numpy as np
from omfit_classes.omfit_mds import OMFITmds, OMFITmdsValue
from omfit_classes.utils_base import compare_version
from omfit_classes.omfit_omas_utils import ensure_consistent_experiment_info

# noinspection PyBroadException
try:
    import omas

    imas_version = omas.ODS().imas_version
except Exception:
    imas_version = '0.0'

__all__ = []


# Decorators
def _id(obj):
    """Trivial decorator as an alternative to make_available()"""
    return obj


def make_available(f):
    """Decorator for listing a function in __all__ so it will be readily available in other scripts"""
    __all__.append(f.__name__)
    return f


def make_available_if(condition):
    """
    Make available only if a condition is met

    :param condition: bool
        True: make_available decorator is used and function is added to __all__ to be readily available
        False: _id decorator is used, function is not added to __all__, and it cannot be accessed quite so easily
            Some functions inspect __all__ to determine which hardware systems are available
    """
    if condition:
        return make_available
    return _id


# Data inspection utilities
@make_available
def find_active_kstar_probes(shot, allowed_probes=None):
    """
    Serves LP functions by identifying active probes (those that have actual data saved) for a given shot

    Sorry, I couldn't figure out how to do this with a server-side loop over all
    the probes, so we have to loop MDS calls on the client side. At least I
    resampled to speed that part up.

    This could be a lot faster if I could figure out how to make the GETNCI
    commands work on records of array signals.

    :param shot: int

    :param allowed_probes: int array
        Restrict the search to a certain range of probe numbers to speed things up

    :return: list of ints
    """
    print(f'Searching for active KSTAR probes in shot {shot}...')
    omv_kw = dict(server='KSTAR', shot=shot, treename='KSTAR')
    lp_paths_unfiltered = OMFITmdsValue(TDI=r'GETNCI("\\TOP.ELECTRON.ELPA:EP*", "MINPATH")', **omv_kw).data()
    lp_paths = [lpp.strip() for lpp in lp_paths_unfiltered if re.match('[0-9]', lpp.strip().split(':')[-1][2:])]
    print(f'{len(lp_paths)} possible probes found')
    lp_nums = np.array([int(lpp.strip().split(':')[-1][2:]) for lpp in lp_paths])
    idx = lp_nums.argsort()
    lp_nums = lp_nums[idx]
    lp_paths = np.array(lp_paths)[idx]
    if allowed_probes is None:
        acceptable_probe_numbers = lp_nums
        acceptable_paths = lp_paths
    else:
        acceptable_probe_numbers = np.array([lpn for lpn in lp_nums if lpn in allowed_probes])
        acceptable_paths = np.array([lp_paths[i] for i in range(len(lp_nums)) if lp_nums[i] in allowed_probes])
        print(f'{len(acceptable_paths)} candidates remain after limiting search to only some specific probes.')
    valid = [None] * len(acceptable_paths)
    i = wd = tc = 0
    for i in range(len(acceptable_paths)):
        ascii_progress_bar(
            i,
            0,
            len(acceptable_paths),
            mess=f'Checking KSTAR#{shot} LP probes for valid data {lp_paths[i]} valid={valid[i]}; '
            f'probes with valid data so far = {wd}/{tc}',
        )
        valid[i] = OMFITmdsValue(TDI=f'resample({acceptable_paths[i]}.FOO, 0, 1, 1)', **omv_kw).check()
        wd = np.sum(np.array(valid).astype(bool))
        tc = len([a for a in valid if a is not None])
    ascii_progress_bar(
        i + 1,
        0,
        len(acceptable_paths),
        mess=f'Checked KSTAR#{shot} LP probes for valid data and found {wd} out of {tc} probes have valid data.',
    )
    print()
    return acceptable_probe_numbers[np.array(valid).astype(bool)]


# Data loading tools
# noinspection PyBroadException
@make_available_if(compare_version(imas_version, '3.25.0') >= 0)  # Might actually be 3.24.X. Close enough for now.
def load_data_langmuir_probes_kstar(ods, shot, probes=None, allowed_probes=None, tstart=0, tend=10, dt=0.0002, overwrite=False):
    """
    Downloads LP probe data from MDSplus and loads them to the ODS

    :param ods: ODS instance

    :param shot: int

    :param probes: int array-like [optional]
        Integer array of KSTAR probe numbers.
        If not provided, find_active_kstar_probes() will be used.

    :param allowed_probes: int array-like [optional]
        Passed to find_active_kstar_probes(), if applicable.
        Improves speed by limiting search to a specific range of probe numbers.

    :param tstart: float
        Time to start resample (s)

    :param tend: float
        Time to end resample (s)
        Set to <= tstart to disable resample

    :param dt: float
        Resample interval (s)
        Set to 0 to disable resample

    :param overwrite: bool
        Download and write data even if they already are present in the ODS.

    :return: ODS instance
        The data are added in-place, so catching the return is probably unnecessary.
    """

    # Make sure the ODS is for the right device/shot
    device = 'KSTAR'
    ensure_consistent_experiment_info(ods, device, shot)

    # Make sure the ODS already has probe positions in it
    try:
        assert is_numeric(ods['langmuir_probes.embedded[0].position.r'])
        assert ods['langmuir_probes.embedded[0].name'] == '1'
    except Exception:
        setup_langmuir_probes_hardware_description_kstar(ods, shot)

    # Select probes to gather
    probes = probes or find_active_kstar_probes(shot, allowed_probes=allowed_probes)

    do_resample = (dt > 0) and (tend > tstart)

    i = 0
    for i, probe in enumerate(probes):
        p_idx = probe - 1
        tdi = rf'\TOP.ELECTRON.ELPA:EP{probe}.FOO'  # Sorry, there doesn't seem to be a pointname for uncertainty.

        ascii_progress_bar(i, 0, len(probes), mess=f'Loading {device}#{shot} LP probe data ({probe}, {tdi})')

        if not overwrite:
            try:
                _ = ods['langmuir_probes.embedded'][p_idx]['time']
                _ = ods['langmuir_probes.embedded'][p_idx]['ion_saturation_current.data']
            except Exception:
                pass
            else:
                printd(f'Skipping probe {probe} with index {p_idx} because it already has data')
                continue

        if do_resample:
            tdi = f'resample({tdi}, {tstart}, {tend}, {dt})'
        m = OMFITmdsValue(server=device, shot=shot, treename='KSTAR', TDI=tdi)
        if m.check():
            # TODO: deal with calibrations and units; this is definitely a real signal that looks like it's
            # proportional to A, but that doesn't mean a factor isn't missing.
            ods['langmuir_probes.embedded'][p_idx]['time'] = m.dim_of(0)
            ods['langmuir_probes.embedded'][p_idx]['ion_saturation_current.data'] = m.data()  # Just nominal values

    ascii_progress_bar(i + 1, 0, len(probes), mess=f'Loaded {device}#{shot} LP probe data')

    return ods


# Hardware descriptor functions
@make_available_if(compare_version(imas_version, '3.25.0') >= 0)  # Might actually be 3.24.X. Close enough for now.
def setup_langmuir_probes_hardware_description_kstar(ods, shot):
    """
    Load KSTAR Langmuir probe locations into an ODS

    :param ods: ODS instance

    :param shot: int

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    import MDSplus

    # Is it okay to try this?
    if compare_version(ods.imas_version, '3.25.0') < 0:
        printe('langmuir_probes.embedded requires a newer version of IMAS. It was added by 3.25.0.')
        printe('ABORTED setup_langmuir_probes_hardware_description_kstar due to old IMAS version.')
        return {}

    # Reused items; set once
    ucf = 1e-3  # Unit Conversion Factor: mm to m
    omv_kw = dict(server='KSTAR', shot=shot, treename='KSTAR')
    base = r'\\TOP.ELECTRON.ELPA:EP*'

    # Get probe numbers from paths in MDSplus
    lp_paths_unfiltered = OMFITmdsValue(TDI=rf'GETNCI("{base}", "MINPATH")', **omv_kw).data()
    lp_paths = [lpp for lpp in lp_paths_unfiltered if re.match('[0-9]', lpp.strip().split(':')[-1][2:])]
    lp_numbers = np.array([int(lpp.strip().split(':')[-1][2:]) for lpp in lp_paths])
    nprobe = len(lp_numbers)

    # Get probe coordinates
    r = OMFITmdsValue(TDI=rf'GETNCI("{base}.R", "RECORD")', **omv_kw).data()
    z = OMFITmdsValue(TDI=rf'GETNCI("{base}.Z", "RECORD")', **omv_kw).data()
    phi = OMFITmdsValue(TDI=rf'GETNCI("{base}.TOR_ANGLE", "RECORD")', **omv_kw).data()
    side_area = OMFITmdsValue(TDI=rf'GETNCI("{base}.NA", "RECORD")', **omv_kw).data()

    # Verify data integrity
    assert len(r) == nprobe, f'Arrays should be length {nprobe} (based on probe#), but R has length {len(r)}'
    assert len(z) == nprobe, f'Arrays should be length {nprobe} (based on probe#), but Z has length {len(z)}'
    assert len(phi) == nprobe, f'Arrays should be length {nprobe} (based on probe#), but phi has length {len(phi)}'
    assert len(side_area) == nprobe, f'Expected len(side_area)={nprobe} (based on probe#), not {len(side_area)}'

    # Sort
    idx = lp_numbers.argsort()
    lp_numbers = lp_numbers[idx]
    r = r[idx]
    z = z[idx]
    phi = phi[idx]
    side_area = side_area[idx]
    # Only the side area is recorded in MDSplus. If it matches data I got from Jun-Gyo BAK, then assume
    # that the top area also matches. Otherwise, we don't know the top area.
    top_area = np.empty(nprobe)
    top_area[:] = np.NaN
    top_area[np.isclose(side_area, 4.088e-6, atol=1e-12)] = 2.827e-5  # m^2

    # Load
    for i in range(nprobe):
        ods['langmuir_probes.embedded'][i]['position.r'] = r[i] * ucf
        ods['langmuir_probes.embedded'][i]['position.z'] = z[i] * ucf
        ods['langmuir_probes.embedded'][i]['position.phi'] = phi[i] * np.pi / 180.0
        # Is having the surface area without the B-field confusing? One cannot just divide by this area
        # to get J_sat; additional transformations are needed.
        ods['langmuir_probes.embedded'][i]['surface_area'] = top_area[i]
        # TODO: Add side surface area, too, if that gets added (see https://jira.iter.org/browse/IMAS-3311)
        ods['langmuir_probes.embedded'][i]['name'] = lp_numbers[i]

    # It should be possible to add surface area as well, but information about angles is needed.

    # From Jun-Gyo BAK at KSTAR:
    # 1. The projected area (of the probe tip) for the normal incidence ( 90 degree )
    # S_1= 2.827 E-5 m^2
    # 2. The projected area for incidence angle of 0 degree
    # S_2= 4.088 E-6 m^2
    # S ~ (S_1*sin(theta) + S_2*cos(theta))
    # Here, theta (incidence angle ) was assumed as 5 degree
    # Thus, J_sat = I_sat/S

    # So, we need the angle between the probe tip and the field.
    # S_2 is recorded in MDSplus as NA. Any probe that has NA = 4.088e-06 m^2
    # should be safe to assume has S_1 = 2.827e-5 m^2, S_2 = 4.088e-6 m^2.
    # Add to langmuir_probes.embedded[:].surface_area

    return {}
